"""

merge_detections.py

Merge high-confidence detections from one or more results files into another
file.  Typically used to combine results from MDv5b and/or MDv4 and/or MDv5a
into a "primary" results file from MDv5a or MDv1000-redwood.

Detection categories must be the same in both files; if you want to first remap
one file's category mapping to be the same as another's, see remap_detection_categories.

If you want to literally merge two .json files, see combine_batch_outputs.py.

"""

#%% Constants and imports

import argparse
import sys
import json
import os

from tqdm import tqdm

from megadetector.utils.ct_utils import get_iou
from megadetector.utils.ct_utils import write_json

from megadetector.postprocessing.classification_postprocessing import \
    merge_classification_categories


#%% Structs

class MergeDetectionsOptions:
    """
    Class defining options for merge_detections().
    """

    def __init__(self):

        #: Maximum detection size to include in the merged output
        self.max_detection_size = 1.01

        #: Minimum detection size to include in the merged output
        self.min_detection_size = 0

        #: Exclude detections whose confidence in the source file(s) is less
        #: than this.  Should have the same length as the number of source files.
        self.source_confidence_thresholds = [0.05]

        #: Don't bother merging into target images if there is a similar detection
        #: above this threshold (or if there is *any* detection above this threshold,
        #: and merge_empty_only is True)
        self.target_confidence_threshold = 0.2

        #: If you want to merge only certain categories, specify one
        #: (but not both) of these.  These are category IDs, not names.
        self.categories_to_include = None

        #: If you want to merge only certain categories, specify one
        #: (but not both) of these.  These are category IDs, not names.
        self.categories_to_exclude = None

        #: Only merge detections into images that have *no* above-threshold
        #: detections in the target results file.
        self.merge_empty_only = False

        #: IoU threshold above which two detections are considered the same
        self.iou_threshold = 0.65

        #: Error if this is False and the output file exists
        self.overwrite = False

        #: Add a 'merged' field (value True) to all merged detections.
        #:
        #: Only used for debugging.
        self.mark_copied_detections = False


#%% Main function

def merge_detections(source_files,
                     target_file,
                     output_file=None,
                     options=None):
    """
    Merge high-confidence detections from one or more results files into another
    file.  Typically used to combine results from, e.g., MDv5a and/or MDv5b into a
    "primary" results file from MDv1000-redwood.

    [source_files] (a list of files or a single filename) specifies the set of
    results files that will be merged into [target_file].  The difference between a
    "source file" and the "target file" is that if no merging is necessary, either because
    two boxes are nearly identical or because merge_only_empty is True and the target
    file already has above-threshold detection for an image+category, the output file gets
    the results of the "target" file.  I.e., the "target" file wins all ties.

    Merges classification categories based on category names (ignoring descriptions).

    The results are written to [output_file].

    Args:
        source_files (list of str): list of files to merge into the results in [target_file]
        target_file (str): filename that is treated as the primary source of results
        output_file (str, optional): file to which we should write merged results
        options (MergeDetectionsOptions, optional): see MergeDetectionsOptions

    Returns:
        dict in MegaDetector format, containing merged results
    """

    if isinstance(source_files,str):
        source_files = [source_files]

    if options is None:
        options = MergeDetectionsOptions()

    if (output_file is not None) and (not options.overwrite) and (os.path.isfile(output_file)):
        print('File {} exists, bypassing merge'.format(output_file))
        return

    assert not ((options.categories_to_exclude is not None) and \
                (options.categories_to_include is not None)), \
                'categories_to_include and categories_to_exclude are mutually exclusive'

    if options.categories_to_exclude is not None:
        options.categories_to_exclude = [int(c) for c in options.categories_to_exclude]

    if options.categories_to_include is not None:
        options.categories_to_include = [int(c) for c in options.categories_to_include]

    assert len(source_files) == len(options.source_confidence_thresholds), \
        '{} source files provided, but {} source confidence thresholds provided'.format(
            len(source_files),len(options.source_confidence_thresholds))

    for fn in source_files:
        assert os.path.isfile(fn), 'Could not find source file {}'.format(fn)

    assert os.path.isfile(target_file)

    with open(target_file,'r') as f:
        output_data = json.load(f)

    print('Loaded results for {} images'.format(len(output_data['images'])))

    fn_to_image = {}

    # im = output_data['images'][0]
    for im in output_data['images']:
        fn_to_image[im['file']] = im

    if 'detections_transferred_from' not in output_data['info']:
        output_data['info']['detections_transferred_from'] = []

    if 'detector' not in output_data['info']:
        output_data['info']['detector'] = 'MDv4 (assumed)'

    detection_categories_raw = output_data['detection_categories'].keys()

    # Determine whether we should be processing all categories, or just a subset
    # of categories.
    detection_categories = []

    if options.categories_to_exclude is not None:
        for c in detection_categories_raw:
            if int(c) not in options.categories_to_exclude:
                detection_categories.append(c)
            else:
                print('Excluding category {}'.format(c))
    elif options.categories_to_include is not None:
        for c in detection_categories_raw:
            if int(c) in options.categories_to_include:
                print('Including category {}'.format(c))
                detection_categories.append(c)
    else:
        detection_categories = detection_categories_raw

    # For each source file...
    #
    # i_source_file = 0; source_file = source_files[i_source_file]
    for i_source_file,source_file in enumerate(source_files):

        print('Processing detections from file {}'.format(source_file))

        with open(source_file,'r') as f:
            source_data = json.load(f)

        # If necessary, map all the classification categories in the source file
        # into the space of the target file
        #
        # This will no-op if neither file has classification categories.
        source_data = \
            merge_classification_categories(target_file=output_data,
                                            source_file=source_data,
                                            output_file=None)

        # If classification data is present in "target_file", merge_classification_categories
        # guarantees compatible classification categories in source_data
        if 'classification_categories' in output_data:
            assert 'classification_categories' in source_data, 'Category merge error A'
            assert len(source_data['classification_categories']) >= \
                   len(output_data['classification_categories']), 'Category merge error B'
            for category_id in output_data['classification_categories']:
                assert category_id in source_data['classification_categories']
                assert source_data['classification_categories'][category_id] == \
                       output_data['classification_categories'][category_id], 'Category merge error C'
            output_data['classification_categories'] = source_data['classification_categories']
            if 'classification_category_descriptions' in output_data:
                output_data['classification_category_descriptions'] = \
                    source_data['classification_category_descriptions']

        if 'detector' in source_data['info']:
            source_detector_name = source_data['info']['detector']
        else:
            source_detector_name = os.path.basename(source_file)

        output_data['info']['detections_transferred_from'].append(os.path.basename(source_file))
        output_data['info']['detector'] = output_data['info']['detector'] + ' + ' + source_detector_name

        assert source_data['detection_categories'] == output_data['detection_categories'], \
            'Cannot merge files with different detection category maps'

        source_confidence_threshold = options.source_confidence_thresholds[i_source_file]

        # For each image...
        #
        # source_im = source_data['images'][0]
        for source_im in tqdm(source_data['images']):

            image_filename = source_im['file']

            assert image_filename in fn_to_image, \
                'Image {} not in target image set'.format(image_filename)
            target_im = fn_to_image[image_filename]

            if 'detections' not in source_im or source_im['detections'] is None:
                continue

            if 'detections' not in target_im or target_im['detections'] is None:
                continue

            source_detections_this_image = source_im['detections']
            target_detections_this_image = target_im['detections']

            detections_to_transfer = []

            # For each detection category...
            #
            # detection_category = list(detection_categories)[0]
            for detection_category in detection_categories:

                target_detections_this_category = \
                    [det for det in target_detections_this_image if det['category'] == \
                     detection_category]

                max_target_confidence_this_category = 0.0

                if len(target_detections_this_category) > 0:
                    max_target_confidence_this_category = max([det['conf'] for \
                      det in target_detections_this_category])

                # If we have a valid detection in the target file, and we're only merging
                # into images that have no detections at all, we don't need to review the individual
                # detections in the source file.
                if options.merge_empty_only and \
                    (max_target_confidence_this_category >= options.target_confidence_threshold):
                    continue

                source_detections_this_category_raw = [det for det in \
                  source_detections_this_image if det['category'] == detection_category]

                # Boxes are x/y/w/h
                # source_sizes = [det['bbox'][2]*det['bbox'][3] for det in source_detections_this_category_raw]

                # Only look at source boxes within the size range
                source_detections_this_category_filtered = [
                    det for det in source_detections_this_category_raw if \
                        (det['bbox'][2]*det['bbox'][3] <= options.max_detection_size) and \
                        (det['bbox'][2]*det['bbox'][3] >= options.min_detection_size) \
                        ]

                # For each detection in this category in the source image...
                #
                # det = source_detections_this_category_filtered[0]
                for det in source_detections_this_category_filtered:

                    if det['conf'] >= source_confidence_threshold:

                        # Check only whole images
                        if options.merge_empty_only:

                            # We verified this above, asserting here for clarity
                            assert max_target_confidence_this_category < options.target_confidence_threshold
                            det['transferred_from'] = source_detector_name
                            detections_to_transfer.append(det)

                        # Check individual detections
                        else:

                            # Does this source detection match any existing above-threshold
                            # target category detections?
                            matches_existing_box = False

                            # target_detection = target_detections_this_category[0]
                            for target_detection in target_detections_this_category:

                                if (target_detection['conf'] >= options.target_confidence_threshold) \
                                   and \
                                   (get_iou(det['bbox'],target_detection['bbox']) >= options.iou_threshold):
                                    matches_existing_box = True
                                    break

                            if (not matches_existing_box):
                                det['transferred_from'] = source_detector_name
                                detections_to_transfer.append(det)

                    # ...if this source detection is above the confidence threshold

                # ...for each source detection within category

            # ...for each detection category

            if len(detections_to_transfer) > 0:

                # print('Adding {} detections to image {}'.format(len(detections_to_transfer),image_filename))
                detections = fn_to_image[image_filename]['detections']
                detections.extend(detections_to_transfer)

                if options.mark_copied_detections:
                    for det in detections_to_transfer:
                        det['merged'] = True

                # Update the max_detection_conf field (if present)
                if 'max_detection_conf' in fn_to_image[image_filename]:
                    fn_to_image[image_filename]['max_detection_conf'] = \
                        max([d['conf'] for d in detections])

            # ...if we have any detections to transfer

        # ...for each image

    # ...for each source file

    if output_file is not None:
        write_json(output_file,output_data)
        print('Saved merged results to {}'.format(output_file))

    return output_data

# ...def merge_detections(...)


#%% Command-line driver

def main():
    """
    Command-line driver for merge_detections.py
    """

    default_options = MergeDetectionsOptions()

    parser = argparse.ArgumentParser(
        description='Merge detections from one or more MegaDetector results files into an existing results file')
    parser.add_argument(
        'source_files',
        nargs='+',
        help='Path to source .json file(s) to merge from')
    parser.add_argument(
        'target_file',
        help='Path to a .json file to merge detections into')
    parser.add_argument(
        'output_file',
        help='Path to output .json results file')
    parser.add_argument(
        '--max_detection_size',
        type=float,
        default=default_options.max_detection_size,
        help='Ignore detections with an area larger than this (as a fraction of ' + \
             'image size) (default {})'.format(
             default_options.max_detection_size))
    parser.add_argument(
        '--min_detection_size',
        default=default_options.min_detection_size,
        type=float,
        help='Ignore detections with an area smaller than this (as a fraction of ' + \
             'image size) (default {})'.format(
             default_options.min_detection_size))
    parser.add_argument(
        '--source_confidence_thresholds',
        nargs='+',
        type=float,
        default=default_options.source_confidence_thresholds,
        help='List of thresholds for each source file (default {}). '.format(
            default_options.source_confidence_thresholds) + \
            'Merge only if the source file\'s detection confidence is higher than its ' + \
            'corresponding threshold.  Should be the same length as the number of source files.')
    parser.add_argument(
        '--target_confidence_threshold',
        type=float,
        default=default_options.target_confidence_threshold,
        help="Do not merge if target file detection confidence is already higher " + \
             "than this (default {})".format(default_options.target_confidence_threshold))
    parser.add_argument(
        '--categories_to_include',
        type=int,
        nargs='+',
        default=None,
        help='List of numeric detection category IDs to include')
    parser.add_argument(
        '--categories_to_exclude',
        type=int,
        nargs='+',
        default=None,
        help='List of numeric detection categories to exclude')
    parser.add_argument(
        '--merge_empty_only',
        action='store_true',
        help='Ignore individual detections and only merge images for which the target ' + \
             'file contains no detections')
    parser.add_argument(
        '--iou_threshold',
        type=float,
        default=default_options.iou_threshold,
        help='Sets the minimum IoU for a source detection to be considered the same as ' + \
             'a target detection (default {})'.format(default_options.iou_threshold))

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    options = MergeDetectionsOptions()
    options.max_detection_size = args.max_detection_size
    options.min_detection_size = args.min_detection_size
    options.source_confidence_thresholds = args.source_confidence_thresholds
    options.target_confidence_threshold = args.target_confidence_threshold
    options.categories_to_include = args.categories_to_include
    options.categories_to_exclude = args.categories_to_exclude
    options.merge_empty_only = args.merge_empty_only
    options.iou_threshold = args.iou_threshold

    merge_detections(args.source_files, args.target_file, args.output_file, options)

if __name__ == '__main__':
    main()



"""

extract_frames_from_video.py

Extracts frames from a source video or folder of videos and writes those frames to jpeg files.
For single videos, writes frame images to the destination folder.  For folders of videos, creates
subfolders in the destination folder (one per video) and writes frame images to those subfolders.

"""

#%% Constants and imports

import argparse
import inspect
import json
import os
import sys

from megadetector.detection.video_utils import \
    video_to_frames, video_folder_to_frames, is_video_file


#%% Options class

class FrameExtractionOptions:
    """
    Parameters controlling the behavior of extract_frames().
    """

    def __init__(self):

        #: Number of workers to use for parallel processing
        self.n_workers = 1

        #: Use threads for parallel processing
        self.parallelize_with_threads = False

        #: JPEG quality for extracted frames
        self.quality = 80

        #: Maximum width for extracted frames (defaults to None)
        self.max_width = None

        #: Enable additional debug output
        self.verbose = False

        #: Sample every Nth frame starting from the first frame; if this is None
        #: or 1, every frame is extracted.  If this is a negative value, it's interpreted
        #: as a sampling rate in seconds, which is rounded to the nearest frame sampling
        #: rate. Mutually exclusive with detector_output_file.
        self.frame_sample = None

        #: Path to MegaDetector .json output file. When specified, extracts frames
        #: referenced in this file.  Mutually exclusive with frame_sample. [source]
        #: must be a folder when this is specified.
        self.detector_output_file = None

    # ...def __init__(...)

# ...class FrameExtractionOptions


#%% Core functions

def extract_frames(source, destination, options=None):
    """
    Extracts frames from a video or folder of videos.

    Args:
        source (str): path to a single video file or folder of videos
        destination (str): folder to write frame images to (will be created if it doesn't exist)
        options (FrameExtractionOptions, optional): parameters controlling frame extraction

    Returns:
        tuple: for single videos, returns (list of frame filenames, frame rate).
               for folders, returns (list of lists of frame filenames, list of frame rates, list
               of video filenames)
    """

    if options is None:
        options = FrameExtractionOptions()

    # Validate inputs
    if not os.path.exists(source):
        raise ValueError('Source path {} does not exist'.format(source))

    if os.path.abspath(source) == os.path.abspath(destination):
        raise ValueError('Source and destination cannot be the same')

    # Create destination folder if it doesn't exist
    os.makedirs(destination, exist_ok=True)

    # Determine whether source is a file or folder
    source_is_file = os.path.isfile(source)

    if source_is_file:

        # Validate that source is a video file
        if not is_video_file(source):
            raise ValueError('Source file {} is not a video file'.format(source))

        # detector_output_file requires source to be a folder
        if options.detector_output_file is not None:
            raise ValueError('detector_output_file option requires source to be a folder, not a file')

        # Extract frames from single video
        return video_to_frames(input_video_file=source,
                               output_folder=destination,
                               overwrite=True,
                               every_n_frames=options.frame_sample,
                               verbose=options.verbose,
                               quality=options.quality,
                               max_width=options.max_width,
                               allow_empty_videos=True)

    else:

        frames_to_extract = None
        relative_paths_to_process = None

        # Handle detector output file
        if options.detector_output_file is not None:
            frames_to_extract, relative_paths_to_process = _parse_detector_output(
                options.detector_output_file, source, options.verbose)
            options.frame_sample = None

        return video_folder_to_frames(input_folder=source,
                                      output_folder_base=destination,
                                      recursive=True,
                                      overwrite=True,
                                      n_threads=options.n_workers,
                                      every_n_frames=options.frame_sample,
                                      verbose=options.verbose,
                                      parallelization_uses_threads=options.parallelize_with_threads,
                                      quality=options.quality,
                                      max_width=options.max_width,
                                      frames_to_extract=frames_to_extract,
                                      relative_paths_to_process=relative_paths_to_process,
                                      allow_empty_videos=True)

# ...def extract_frames(...)


def _parse_detector_output(detector_output_file, source_folder, verbose=False):
    """
    Parses a MegaDetector .json output file and returns frame extraction information.

    Args:
        detector_output_file (str): path to MegaDetector .json output file
        source_folder (str): folder containing the source videos
        verbose (bool, optional): enable additional debug output

    Returns:
        tuple: (frames_to_extract_dict, relative_paths_to_process) where:
               - frames_to_extract_dict maps relative video paths to lists of frame numbers
               - relative_paths_to_process is a list of relative video paths to process
    """

    print('Parsing detector output file: {}'.format(detector_output_file))

    # Load the detector results
    with open(detector_output_file, 'r') as f:
        detector_results = json.load(f)

    if 'images' not in detector_results:
        raise ValueError('Detector output file does not contain "images" field')

    images = detector_results['images']
    frames_to_extract_dict = {}
    video_files_in_results = set()

    for image_entry in images:

        file_path = image_entry['file']

        # Skip non-video files
        if not is_video_file(file_path):
            if verbose:
                print('Skipping non-video file {}'.format(file_path))
            continue

        # Check whether video file exists in source folder
        full_video_path = os.path.join(source_folder, file_path)
        if not os.path.isfile(full_video_path):
            print('Warning: video file {} not found in source folder, skipping'.format(file_path))
            continue

        video_files_in_results.add(file_path)

        # Determine which frames to extract for this video
        frames_for_this_video = []

        if 'frames_processed' in image_entry:
            # Use the frames_processed field if available
            frames_for_this_video = image_entry['frames_processed']
            if verbose:
                print('Video {}: using frames_processed field with {} frames'.format(
                    file_path, len(frames_for_this_video)))
        else:
            # Extract frames from detections
            if ('detections' in image_entry) and (image_entry['detections'] is not None):
                frame_numbers = set()
                for detection in image_entry['detections']:
                    if 'frame_number' in detection:
                        frame_numbers.add(detection['frame_number'])
                frames_for_this_video = sorted(list(frame_numbers))
                if verbose:
                    print('Video {}: extracted {} unique frame numbers from detections'.format(
                        file_path, len(frames_for_this_video)))

        if len(frames_for_this_video) > 0:
            frames_to_extract_dict[file_path] = frames_for_this_video

    # ...for each image/video in this file

    relative_paths_to_process = sorted(list(video_files_in_results))

    print('Found {} videos with frames to extract'.format(len(frames_to_extract_dict)))

    return frames_to_extract_dict, relative_paths_to_process

# ...def _parse_detector_output(...)


#%% Command-line driver

def _args_to_object(args, obj):
    """
    Copy all fields from a Namespace (i.e., the output from parse_args) to an object.
    Skips fields starting with _.  Does not check existence in the target object.
    """

    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)


def main():
    """
    Command-line driver for extract_frames_from_video
    """

    parser = argparse.ArgumentParser(
        description='Extract frames from videos and save as JPEG files')

    parser.add_argument('source', type=str,
                        help='Path to a single video file or folder containing videos')
    parser.add_argument('destination', type=str,
                        help='Output folder for extracted frames (will be created if it does not exist)')

    parser.add_argument('--n_workers', type=int, default=1,
                        help='Number of workers to use for parallel processing (default: %(default)s)')
    parser.add_argument('--parallelize_with_threads', action='store_true',
                        help='Use threads for parallel processing (default: use processes)')
    parser.add_argument('--quality', type=int, default=80,
                        help='JPEG quality for extracted frames (default: %(default)s)')
    parser.add_argument('--max_width', type=int, default=None,
                        help='Maximum width for extracted frames (default: no resizing)')
    parser.add_argument('--verbose', action='store_true',
                        help='Enable additional debug output')

    # Mutually exclusive group for frame sampling options
    frame_group = parser.add_mutually_exclusive_group()
    frame_group.add_argument('--frame_sample', type=float, default=None,
                             help='Sample every Nth frame starting from the first frame; if this is None or 1, ' +
                                  'every frame is extracted. If this is a negative value, it\'s interpreted as a ' +
                                  'sampling rate in seconds, which is rounded to the nearest frame sampling rate')
    frame_group.add_argument('--detector_output_file', type=str, default=None,
                             help='Path to MegaDetector .json output file. When specified, extracts frames ' +
                                  'referenced in this file. Source must be a folder when this is specified.')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # Convert to an options object
    options = FrameExtractionOptions()
    _args_to_object(args, options)

    # Additional validation
    if options.detector_output_file is not None:
        if not os.path.isfile(options.detector_output_file):
            print('Error: detector_output_file {} does not exist'.format(options.detector_output_file))
            sys.exit(1)

    try:
        result = extract_frames(args.source, args.destination, options)

        if os.path.isfile(args.source):
            frame_filenames, frame_rate = result
            print('Extracted {} frames from {} (frame rate: {:.2f} fps)'.format(
                len(frame_filenames), args.source, frame_rate))
        else:
            frame_filenames_by_video, fs_by_video, video_filenames = result
            total_frames = sum(len(frames) for frames in frame_filenames_by_video)
            print('Processed {} videos, extracted {} total frames'.format(
                len(video_filenames), total_frames))

    except Exception as e:
        print('Error: {}'.format(str(e)))
        sys.exit(1)


if __name__ == '__main__':
    main()



"""

zamba_to_md.py

Convert a labels.csv file produced by Zamba Cloud to a MD results file suitable
for import into Timelapse.

Columns are expected to be:

video_uuid (not used)
original_filename (assumed to be a relative path name)
top_k_label,top_k_probability, for k = 1..N
[category name 1],[category name 2],...
corrected_label

Because the MD results file fundamentally stores detections, what we'll
actually do is create bogus detections that fill the entire image.

There is no special handling of empty/blank categories; because these results are
based on a classifier, rather than a detector (where "blank" would be the absence of
all other categories), "blank" can be queried in Timelapse just like any other class.

"""

#%% Imports and constants

import sys
import argparse

import pandas as pd

from megadetector.utils.ct_utils import write_json


#%% Main function

def zamba_results_to_md_results(input_file,output_file=None):
    """
    Converts the .csv file [input_file] to the MD-formatted .json file [output_file].

    If [output_file] is None, '.json' will be appended to the input file.

    Args:
        input_file (str): the .csv file to convert
        output_file (str, optional): the output .json file (defaults to
            [input_file].json)
    """

    if output_file is None:
        output_file = input_file + '.json'

    df = pd.read_csv(input_file)

    expected_columns = ('video_uuid','corrected_label','original_filename')
    for s in expected_columns:
        assert s in df.columns,\
            'Expected column {} not found, are you sure this is a Zamba results .csv file?'.format(
                s)

    # How many results are included per file?
    assert 'top_1_probability' in df.columns and 'top_1_label' in df.columns
    top_k = 2
    while(True):
        p_string = 'top_' + str(top_k) + '_probability'
        label_string = 'top_' + str(top_k) + '_label'

        if p_string in df.columns:
            assert label_string in df.columns,\
                'Oops, {} is a column but {} is not'.format(
                    p_string,label_string)
            top_k += 1
            continue
        else:
            assert label_string not in df.columns,\
                'Oops, {} is a column but {} is not'.format(
                    label_string,p_string)
            top_k -= 1
            break

    print('Found {} probability column pairs'.format(top_k))

    # Category names start after the fixed columns and the probability columns
    category_names = []
    column_names = list(df.columns)
    first_category_name_index = 0
    while('top_' in column_names[first_category_name_index] or \
          column_names[first_category_name_index] in expected_columns):
        first_category_name_index += 1

    i_column = first_category_name_index
    while( (i_column < len(column_names)) and (column_names[i_column] != 'corrected_label') ):
        category_names.append(column_names[i_column])
        i_column += 1

    print('Found {} categories:\n'.format(len(category_names)))

    for s in category_names:
        print(s)

    info = {}
    info['format_version'] = '1.3'
    info['detector'] = 'Zamba Cloud'
    info['classifier'] = 'Zamba Cloud'

    detection_category_id_to_name = {}
    for category_id,category_name in enumerate(category_names):
        detection_category_id_to_name[str(category_id)] = category_name
    detection_category_name_to_id = {v: k for k, v in detection_category_id_to_name.items()}

    images = []

    # i_row = 0; row = df.iloc[i_row]
    for i_row,row in df.iterrows():

        im = {}
        images.append(im)
        im['file'] = row['original_filename']

        detections = []

        # k = 1
        for k in range(1,top_k+1):
            label = row['top_{}_label'.format(k)]
            confidence = row['top_{}_probability'.format(k)]
            det = {}
            det['category'] = detection_category_name_to_id[label]
            det['conf'] = confidence
            det['bbox'] = [0,0,1.0,1.0]
            detections.append(det)

        im['detections'] = detections

    # ...for each row

    results = {}
    results['info'] = info
    results['detection_categories'] = detection_category_id_to_name
    results['images'] = images

    write_json(output_file,results)

# ...zamba_results_to_md_results(...)


#%% Interactive driver

if False:

    pass

    #%%

    input_file = r"G:\temp\labels-job-b95a4b76-e332-4e17-ab40-03469392d36a-2023-11-04_16-28-50.060130.csv"
    output_file = None
    zamba_results_to_md_results(input_file,output_file)


#%% Command-line driver

def main():
    """
    Command-line driver for zamba_to_md
    """

    parser = argparse.ArgumentParser(
        description='Convert a Zamba-formatted .csv results file to a MD-formatted .json results file')

    parser.add_argument(
        'input_file',
        type=str,
        help='input .csv file')

    parser.add_argument(
        '--output_file',
        type=str,
        default=None,
        help='output .json file (defaults to input file appended with ".json")')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    zamba_results_to_md_results(args.input_file,args.output_file)

if __name__ == '__main__':
    main()



"""

species_lookup.py

Look up species names (common or scientific) in the GBIF and iNaturalist
taxonomies.

Run initialize_taxonomy_lookup() before calling any other function.

"""

#%% Constants and imports

import argparse
import pickle
import shutil
import zipfile
import sys
import os

from collections import defaultdict
from itertools import compress
from tqdm import tqdm
from typing import Any, Dict, List, Mapping, Sequence, Set

import pandas as pd
import numpy as np

from megadetector.utils import url_utils

taxonomy_download_dir = os.path.expanduser('~/taxonomy')

taxonomy_urls = {
    'GBIF': 'https://hosted-datasets.gbif.org/datasets/backbone/current/backbone.zip',
    'iNaturalist': 'https://www.inaturalist.org/taxa/inaturalist-taxonomy.dwca.zip'
}

files_to_unzip = {
    'GBIF': ['Taxon.tsv', 'VernacularName.tsv'],
    'iNaturalist': ['taxa.csv','VernacularNames-english.csv']
}

# As of 2025.06.24:
#
# GBIF: 950MB zipped, 2.3GB of relevant content unzipped
# iNat: 71MB zipped, 415MB of relevant content unzipped

os.makedirs(taxonomy_download_dir, exist_ok=True)
for taxonomy_name in taxonomy_urls:
    taxonomy_dir = os.path.join(taxonomy_download_dir, taxonomy_name)
    os.makedirs(taxonomy_dir, exist_ok=True)

serialized_structures_file = os.path.join(taxonomy_download_dir,
                                          'serialized_taxonomies.p')

# These are un-initialized globals that must be initialized by
# the initialize_taxonomy_lookup() function below.
inat_taxonomy = None # : pd.DataFrame
gbif_taxonomy = None # : pd.DataFrame
gbif_common_mapping = None # : pd.DataFrame
inat_taxon_id_to_row = None # : Dict[np.int64, int]
gbif_taxon_id_to_row = None # : Dict[np.int64, int]
inat_taxon_id_to_vernacular = None # : Dict[np.int64, Set[str]]
inat_vernacular_to_taxon_id = None # : Dict[str, np.int64]
inat_taxon_id_to_scientific = None # : Dict[np.int64, Set[str]]
inat_scientific_to_taxon_id = None # : Dict[str, np.int64]
gbif_taxon_id_to_vernacular = None # : Dict[np.int64, Set[str]]
gbif_vernacular_to_taxon_id = None # : Dict[str, np.int64]
gbif_taxon_id_to_scientific = None # : Dict[np.int64, Set[str]]
gbif_scientific_to_taxon_id = None # : Dict[str, np.int64]


#%% Functions

# Initialization function

def initialize_taxonomy_lookup(force_init=False):
    """
    Initialize this module by doing the following:

    * Downloads and unzips the current GBIF and iNat taxonomies if necessary
      (only unzips what's necessary, but does not delete the original zipfiles)
    * Builds a bunch of dictionaries and tables to facilitate lookup
    * Serializes those tables via pickle
    * Skips all of the above if the serialized pickle file already exists

    Args:
        force_init (bool, optional): force re-download and parsing of the source .zip files,
            even if the cached .p file already exists
    """

    #%%

    global inat_taxonomy,\
        gbif_taxonomy,\
        gbif_common_mapping,\
        inat_taxon_id_to_row,\
        gbif_taxon_id_to_row,\
        inat_taxon_id_to_vernacular,\
        inat_vernacular_to_taxon_id,\
        inat_taxon_id_to_scientific,\
        inat_scientific_to_taxon_id,\
        gbif_taxon_id_to_vernacular,\
        gbif_vernacular_to_taxon_id,\
        gbif_taxon_id_to_scientific,\
        gbif_scientific_to_taxon_id


    #%% Load serialized taxonomy info if we've already saved it

    if (not force_init) and (inat_taxonomy is not None):
        print('Skipping taxonomy re-init')
        return

    if (not force_init) and (os.path.isfile(serialized_structures_file)):

        print(f'De-serializing taxonomy data from {serialized_structures_file}')

        with open(serialized_structures_file, 'rb') as f:
            structures_to_serialize = pickle.load(f)

        inat_taxonomy,\
        gbif_taxonomy,\
        gbif_common_mapping,\
        inat_taxon_id_to_row,\
        gbif_taxon_id_to_row,\
        inat_taxon_id_to_vernacular,\
        inat_vernacular_to_taxon_id,\
        inat_taxon_id_to_scientific,\
        inat_scientific_to_taxon_id,\
        gbif_taxon_id_to_vernacular,\
        gbif_vernacular_to_taxon_id,\
        gbif_taxon_id_to_scientific,\
        gbif_scientific_to_taxon_id = structures_to_serialize

        return


    #%% Download and unzip taxonomy files

    # taxonomy_name = list(taxonomy_urls.items())[0][0]; zip_url = list(taxonomy_urls.items())[0][1]
    for taxonomy_name, zip_url in taxonomy_urls.items():

        need_to_download = False

        if force_init:
            need_to_download = True

        # Don't download the zipfile if we've already unzipped what we need
        for fn in files_to_unzip[taxonomy_name]:
            target_file = os.path.join(
                taxonomy_download_dir, taxonomy_name, fn)
            if not os.path.isfile(target_file):
                need_to_download = True
                break
        if not need_to_download:
            print(f'Bypassing download of {taxonomy_name}, all files available')
            continue

        zipfile_path = os.path.join(
            taxonomy_download_dir, zip_url.split('/')[-1])

        # Bypasses download if the file exists already (unless force_init is set)
        url_utils.download_url(
            zip_url, os.path.join(zipfile_path),
            progress_updater=url_utils.DownloadProgressBar(),
            verbose=True,force_download=force_init)

        # Unzip the files we need
        files_we_need = files_to_unzip[taxonomy_name]

        with zipfile.ZipFile(zipfile_path, 'r') as zipH:

            for fn in files_we_need:
                print('Unzipping {}'.format(fn))
                target_file = os.path.join(
                    taxonomy_download_dir, taxonomy_name, os.path.basename(fn))

                if (not force_init) and (os.path.isfile(target_file)):
                    print(f'Bypassing unzip of {target_file}, file exists')
                else:
                    os.makedirs(os.path.basename(target_file),exist_ok=True)
                    with zipH.open(fn) as zf, open(target_file, 'wb') as f:
                        shutil.copyfileobj(zf, f)

            # ...for each file that we need from this zipfile

    # ...for each taxonomy


    #%% Create dataframes from each of the taxonomy/vernacular files

    # Load iNat taxonomy
    inat_taxonomy_file = os.path.join(taxonomy_download_dir, 'iNaturalist', 'taxa.csv')
    print('Loading iNat taxonomy from {}'.format(inat_taxonomy_file))
    inat_taxonomy = pd.read_csv(inat_taxonomy_file)
    inat_taxonomy['scientificName'] = inat_taxonomy['scientificName'].fillna('').str.strip()

    # Delete columns we won't use.  The "taxonID" column is a non-int version of "ID"
    inat_taxonomy = inat_taxonomy.drop(['identifier', 'taxonID', 'modified', 'references'], axis=1)

    # The "parentNameUsageID" column in inat_taxonomy is a URL, like:
    #
    # https://www.inaturalist.org/taxa/71262
    #
    # Convert this column to be integer-valued, using only the last token of the URL
    inat_taxonomy['parentNameUsageID'] = \
        inat_taxonomy['parentNameUsageID'].str.split('/').str[-1].fillna(0).astype(int)

    # Rename the "id" column to "taxonID"
    inat_taxonomy = inat_taxonomy.rename(columns={'id': 'taxonID'})

    assert 'id' not in inat_taxonomy.columns
    assert 'taxonID' in inat_taxonomy.columns

    # Load iNat common name mapping
    inat_common_mapping_file = os.path.join(taxonomy_download_dir, 'iNaturalist', 'VernacularNames-english.csv')
    inat_common_mapping = pd.read_csv(inat_common_mapping_file)
    inat_common_mapping['vernacularName'] = inat_common_mapping['vernacularName'].fillna('').str.strip()

    inat_common_mapping = inat_common_mapping.drop(['language','locality','countryCode',
                                                    'source','lexicon','contributor','created'], axis=1)
    assert 'id' in inat_common_mapping.columns
    assert 'taxonID' not in inat_common_mapping.columns
    assert 'vernacularName' in inat_common_mapping.columns

    # Load GBIF taxonomy
    gbif_taxonomy_file = os.path.join(taxonomy_download_dir, 'GBIF', 'Taxon.tsv')
    print('Loading GBIF taxonomy from {}'.format(gbif_taxonomy_file))
    gbif_taxonomy = pd.read_csv(gbif_taxonomy_file, sep='\t', encoding='utf-8',on_bad_lines='warn')
    gbif_taxonomy['scientificName'] = gbif_taxonomy['scientificName'].fillna('').str.strip()
    gbif_taxonomy['canonicalName'] = gbif_taxonomy['canonicalName'].fillna('').str.strip()
    gbif_taxonomy['parentNameUsageID'] = gbif_taxonomy['parentNameUsageID'].fillna(-1).astype(int)

    # Remove questionable rows from the GBIF taxonomy
    gbif_taxonomy = gbif_taxonomy[~gbif_taxonomy['taxonomicStatus'].isin(['doubtful', 'misapplied'])]
    gbif_taxonomy = gbif_taxonomy.reset_index()

    gbif_taxonomy = gbif_taxonomy.drop(['datasetID','acceptedNameUsageID','originalNameUsageID',
                                        'scientificNameAuthorship','nameAccordingTo','namePublishedIn',
                                        'taxonomicStatus','nomenclaturalStatus','taxonRemarks'], axis=1)

    assert 'taxonID' in gbif_taxonomy.columns
    assert 'scientificName' in gbif_taxonomy.columns

    # Load GBIF common name mapping
    gbif_common_mapping = pd.read_csv(os.path.join(
        taxonomy_download_dir, 'GBIF', 'VernacularName.tsv'), sep='\t')
    gbif_common_mapping['vernacularName'] = gbif_common_mapping['vernacularName'].fillna('').str.strip()

    # Only keep English mappings
    gbif_common_mapping = gbif_common_mapping.loc[gbif_common_mapping['language'] == 'en']
    gbif_common_mapping = gbif_common_mapping.reset_index()

    gbif_common_mapping = gbif_common_mapping.drop(['language','country','countryCode','sex',
                                                    'lifeStage','source'],axis=1)

    assert 'taxonID' in gbif_common_mapping.columns
    assert 'vernacularName' in gbif_common_mapping.columns


    # Convert everything to lowercase

    def convert_df_to_lowercase(df):
        df = df.applymap(lambda s: s.lower() if isinstance(s, str) else s)
        return df

    inat_taxonomy = convert_df_to_lowercase(inat_taxonomy)
    gbif_taxonomy = convert_df_to_lowercase(gbif_taxonomy)
    gbif_common_mapping = convert_df_to_lowercase(gbif_common_mapping)
    inat_common_mapping = convert_df_to_lowercase(inat_common_mapping)


    ##%% For each taxonomy table, create a mapping from taxon IDs to rows

    inat_taxon_id_to_row = {}
    gbif_taxon_id_to_row = {}

    print('Building iNat taxonID --> row table')
    for i_row, row in tqdm(inat_taxonomy.iterrows(), total=len(inat_taxonomy)):
        taxon_id = row['taxonID']
        assert isinstance(taxon_id, int)
        inat_taxon_id_to_row[taxon_id] = i_row

    print('Building GBIF taxonID --> row table')
    for i_row, row in tqdm(gbif_taxonomy.iterrows(), total=len(gbif_taxonomy)):
        taxon_id = row['taxonID']
        assert isinstance(taxon_id, int)
        gbif_taxon_id_to_row[taxon_id] = i_row


    ##%% Create name mapping dictionaries

    inat_taxon_id_to_vernacular = defaultdict(set)
    inat_vernacular_to_taxon_id = defaultdict(set)
    inat_taxon_id_to_scientific = defaultdict(set)
    inat_scientific_to_taxon_id = defaultdict(set)

    gbif_taxon_id_to_vernacular = defaultdict(set)
    gbif_vernacular_to_taxon_id = defaultdict(set)
    gbif_taxon_id_to_scientific = defaultdict(set)
    gbif_scientific_to_taxon_id = defaultdict(set)


    # Build iNat dictionaries

    print('Building lookup dictionaries for iNat taxonomy')

    # iNat Scientific name mapping

    for i_row, row in tqdm(inat_taxonomy.iterrows(), total=len(inat_taxonomy)):

        taxon_id = row['taxonID']
        assert isinstance(taxon_id,int)

        scientific_name = row['scientificName']
        assert len(scientific_name) > 0

        inat_taxon_id_to_scientific[taxon_id].add(scientific_name)
        inat_scientific_to_taxon_id[scientific_name].add(taxon_id)

    # iNat common name mapping

    inat_taxon_ids_in_vernacular_file_but_not_in_taxa_file = set()

    for i_row, row in tqdm(inat_common_mapping.iterrows(), total=len(inat_common_mapping)):

        taxon_id = row['id']
        assert isinstance(taxon_id,int)

        # This should never happen; we will assert() this at the end of the loop
        if taxon_id not in inat_taxon_id_to_scientific:
            inat_taxon_ids_in_vernacular_file_but_not_in_taxa_file.add(taxon_id)
            continue

        vernacular_name = row['vernacularName']

        assert len(vernacular_name) > 0
        inat_taxon_id_to_vernacular[taxon_id].add(vernacular_name)
        inat_vernacular_to_taxon_id[vernacular_name].add(taxon_id)

    assert len(inat_taxon_ids_in_vernacular_file_but_not_in_taxa_file) == 0


    ##%% Build GBIF dictionaries

    print('Building lookup dictionaries for GBIF taxonomy')

    # GBIF scientific name mapping

    for i_row, row in tqdm(gbif_taxonomy.iterrows(), total=len(gbif_taxonomy)):

        taxon_id = row['taxonID']
        assert isinstance(taxon_id,int)

        # The "canonical name" is the Latin name; the "scientific name"
        # column includes other information.  For example:
        #
        # "scientificName": Schizophoria impressa (Hall, 1843)
        # "canonicalName": Schizophoria impressa
        #
        # Also see:
        #
        # http://globalnames.org/docs/glossary/

        scientific_name = row['canonicalName']

        # This only seems to happen for really esoteric species that aren't
        # likely to apply to our problems, but doing this for completeness.
        if len(scientific_name) == 0:
            scientific_name = row['scientificName']

        assert len(scientific_name) > 0
        gbif_taxon_id_to_scientific[taxon_id].add(scientific_name)
        gbif_scientific_to_taxon_id[scientific_name].add(taxon_id)

    # GBIF common name mapping

    gbif_taxon_ids_in_vernacular_file_but_not_in_taxa_file = set()

    for i_row, row in tqdm(gbif_common_mapping.iterrows(), total=len(gbif_common_mapping)):

        taxon_id = row['taxonID']
        assert isinstance(taxon_id,int)

        # Don't include taxon IDs that were removed from the master table
        if taxon_id not in gbif_taxon_id_to_scientific:
            gbif_taxon_ids_in_vernacular_file_but_not_in_taxa_file.add(taxon_id)
            continue

        vernacular_name = row['vernacularName']

        assert len(vernacular_name) > 0
        gbif_taxon_id_to_vernacular[taxon_id].add(vernacular_name)
        gbif_vernacular_to_taxon_id[vernacular_name].add(taxon_id)

    print('Finished GBIF common --> scientific mapping, failed to map {} of {} taxon IDs'.format(
        len(gbif_taxon_ids_in_vernacular_file_but_not_in_taxa_file),
        len(gbif_common_mapping)
    ))


    ##%% Save everything to file

    structures_to_serialize = [
        inat_taxonomy,
        gbif_taxonomy,
        gbif_common_mapping,
        inat_taxon_id_to_row,
        gbif_taxon_id_to_row,
        inat_taxon_id_to_vernacular,
        inat_vernacular_to_taxon_id,
        inat_taxon_id_to_scientific,
        inat_scientific_to_taxon_id,
        gbif_taxon_id_to_vernacular,
        gbif_vernacular_to_taxon_id,
        gbif_taxon_id_to_scientific,
        gbif_scientific_to_taxon_id
    ]

    print('Serializing to {}...'.format(serialized_structures_file), end='')
    if not os.path.isfile(serialized_structures_file):
        with open(serialized_structures_file, 'wb') as p:
            pickle.dump(structures_to_serialize, p)
    print('...done')


    #%%

# ...def initialize_taxonomy_lookup(...)


def get_scientific_name_from_row(r):
    """
    r: a dataframe that's really a row in one of our taxonomy tables
    """

    if 'canonicalName' in r and len(r['canonicalName']) > 0:
        scientific_name = r['canonicalName']
    else:
        scientific_name = r['scientificName']
    return scientific_name


def taxonomy_row_to_string(r):
    """
    r: a dataframe that's really a row in one of our taxonomy tables
    """

    if 'vernacularName' in r:
        common_string = ' (' + r['vernacularName'] + ')'
    else:
        common_string = ''
    scientific_name = get_scientific_name_from_row(r)

    return r['taxonRank'] + ' ' + scientific_name + common_string


def traverse_taxonomy(matching_rownums: Sequence[int],
                      taxon_id_to_row: Mapping[str, int],
                      taxon_id_to_vernacular: Mapping[str, Set[str]],
                      taxonomy: pd.DataFrame,
                      source_name: str,
                      query: str) -> List[Dict[str, Any]]:
    """
    Given a data frame that's a set of rows from one of our taxonomy tables,
    walks the taxonomy hierarchy from each row to put together a full taxonomy
    tree, then prunes redundant trees (e.g. if we had separate hits for a
    species and the genus that contains that species.)

    Returns a list of dicts:
    [
      {
        'source': 'inat' or 'gbif',
        'taxonomy': [(taxon_id, taxon_rank, scientific_name, [common names])]
      },
      ...
    ]
    """

    # list of dicts: {'source': source_name, 'taxonomy': match_details}
    matching_trees: List[Dict[str, Any]] = []

    # i_match = 0
    for i_match in matching_rownums:

        # list of (taxon_id, taxonRank, scientific name, [vernacular names])
        # corresponding to an exact match and its parents
        match_details = []
        current_row = taxonomy.iloc[i_match]

        # Walk taxonomy hierarchy
        while True:

            taxon_id = current_row['taxonID']
            # sort for determinism
            vernacular_names = sorted(taxon_id_to_vernacular[taxon_id])
            match_details.append((taxon_id, current_row['taxonRank'],
                                  get_scientific_name_from_row(current_row),
                                  vernacular_names))

            if np.isnan(current_row['parentNameUsageID']):
                break
            parent_taxon_id = current_row['parentNameUsageID'].astype('int64')
            if parent_taxon_id not in taxon_id_to_row:
                # This can happen because we remove questionable rows from the
                # GBIF taxonomy
                # print(f'Warning: no row exists for parent_taxon_id {parent_taxon_id},' + \
                #      f'child taxon_id: {taxon_id}, query: {query}')
                break
            i_parent_row = taxon_id_to_row[parent_taxon_id]
            current_row = taxonomy.iloc[i_parent_row]

            # The GBIF taxonomy contains unranked entries
            if current_row['taxonRank'] == 'unranked':
                break

        # ...while there is taxonomy left to walk

        matching_trees.append({'source': source_name,
                               'taxonomy': match_details})

    # ...for each match

    # Remove redundant matches
    b_valid_tree = [True] * len(matching_rownums)
    # i_tree_a = 0; tree_a = matching_trees[i_tree_a]
    for i_tree_a, tree_a in enumerate(matching_trees):

        tree_a_primary_taxon_id = tree_a['taxonomy'][0][0]

        # i_tree_b = 1; tree_b = matching_trees[i_tree_b]
        for i_tree_b, tree_b in enumerate(matching_trees):

            if i_tree_a == i_tree_b:
                continue

            # If tree a's primary taxon ID is inside tree b, discard tree a
            #
            # taxonomy_level_b = tree_b['taxonomy'][0]
            for taxonomy_level_b in tree_b['taxonomy']:
                if tree_a_primary_taxon_id == taxonomy_level_b[0]:
                    b_valid_tree[i_tree_a] = False
                    break

            # ...for each level in taxonomy B

        # ...for each tree (inner)

    # ...for each tree (outer)

    matching_trees = list(compress(matching_trees, b_valid_tree))
    return matching_trees

# ...def traverse_taxonomy()


def get_taxonomic_info(query: str) -> List[Dict[str, Any]]:
    """
    Main entry point: get taxonomic matches from both taxonomies for [query],
    which may be a scientific or common name.
    """

    query = query.strip().lower()
    # print("Finding taxonomy information for: {0}".format(query))

    inat_taxon_ids = set()
    if query in inat_scientific_to_taxon_id:
        inat_taxon_ids |= inat_scientific_to_taxon_id[query]
    if query in inat_vernacular_to_taxon_id:
        inat_taxon_ids |= inat_vernacular_to_taxon_id[query]

    # In GBIF, some queries hit for both common and scientific, make sure we end
    # up with unique inputs
    gbif_taxon_ids = set()
    if query in gbif_scientific_to_taxon_id:
        gbif_taxon_ids |= gbif_scientific_to_taxon_id[query]
    if query in gbif_vernacular_to_taxon_id:
        gbif_taxon_ids |= gbif_vernacular_to_taxon_id[query]

    # If the species is not found in either taxonomy, return None
    if (len(inat_taxon_ids) == 0) and (len(gbif_taxon_ids) == 0):
        return []

    # Both GBIF and iNat have a 1-to-1 mapping between taxon_id and row number
    inat_row_indices = [inat_taxon_id_to_row[i] for i in inat_taxon_ids]
    gbif_row_indices = [gbif_taxon_id_to_row[i] for i in gbif_taxon_ids]

    # Walk both taxonomies
    inat_matching_trees = traverse_taxonomy(
        inat_row_indices, inat_taxon_id_to_row, inat_taxon_id_to_vernacular,
        inat_taxonomy, 'inat', query)
    gbif_matching_trees = traverse_taxonomy(
        gbif_row_indices, gbif_taxon_id_to_row, gbif_taxon_id_to_vernacular,
        gbif_taxonomy, 'gbif', query)

    return gbif_matching_trees + inat_matching_trees

# ...def get_taxonomic_info()


def print_taxonomy_matches(matches, verbose=False):
    """
    Console-friendly printing function to make nicely-indentend trees
    """

    # m = matches[0]
    for m in matches:

        source = m['source']

        # For example: [(9761484, 'species', 'anas platyrhynchos')]
        for i_taxonomy_level in range(0, len(m['taxonomy'])):
            taxonomy_level_info = m['taxonomy'][i_taxonomy_level]
            taxonomy_level = taxonomy_level_info[1]
            name = taxonomy_level_info[2]
            common = taxonomy_level_info[3]

            if i_taxonomy_level > 0:
                print('\t',end='')

            print('{} {} ({})'.format(taxonomy_level, name, common), end='')

            if i_taxonomy_level == 0:
                print(' ({})'.format(source))
            else:
                print('')

            if not verbose:
                break

        # ...for each taxonomy level

    # ...for each match

# ...def print_taxonomy_matches()


#%% Taxonomy functions that make subjective judgements

import unicodedata
import re

def slugify(value: Any, allow_unicode: bool = False) -> str:
    """
    From:
    https://github.com/django/django/blob/master/django/utils/text.py

    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """

    value = str(value)
    value = unicodedata.normalize('NFKC', value)
    if not allow_unicode:
        value = value.encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value.lower()).strip()
    return re.sub(r'[-\s]+', '-', value)


class TaxonomicMatch:

    def __init__(self, scientific_name, common_name, taxonomic_level, source,
                 taxonomy_string, match):
        self.scientific_name = scientific_name
        self.common_name = common_name
        self.taxonomic_level = taxonomic_level
        self.source = source
        self.taxonomy_string = taxonomy_string
        self.match = match

    def __repr__(self):
        return ('TaxonomicMatch('
            f'scientific_name={self.scientific_name}, '
            f'common_name={self.common_name}, '
            f'taxonomic_level={self.taxonomic_level}, '
            f'source={self.source}')


hyphenated_terms = ['crowned', 'backed', 'throated', 'tailed', 'headed', 'cheeked',
                    'ruffed', 'browed', 'eating', 'striped', 'shanked',
                    'fronted', 'bellied', 'spotted', 'eared', 'collared', 'breasted',
                    'necked']

def pop_levels(m, n_levels=1):
    """
    Remove [n_levels] levels from the bottom of the TaxonomicMatch object m, typically used to remove
    silly subgenera.
    """

    v = eval(m.taxonomy_string)
    assert v[0][1] == m.taxonomic_level
    assert v[0][2] == m.scientific_name
    popped_v = v[n_levels:]
    taxonomic_level = popped_v[0][1]
    scientific_name = popped_v[0][2]
    common_name = popped_v[0][3]
    if len(common_name) == 0:
        common_name = ''
    else:
        common_name = common_name[0]
    taxonomy_string = str(popped_v)
    source = m.source
    return TaxonomicMatch(scientific_name=scientific_name,
                          common_name=common_name,
                          taxonomic_level=taxonomic_level,
                          source=source,
                          taxonomy_string=taxonomy_string,
                          match=None)

# ...def pop_levels(...)


def get_preferred_taxonomic_match(query: str, taxonomy_preference = 'inat', retry=True) -> TaxonomicMatch:
    """
    Wrapper for _get_preferred_taxonomic_match, but expressing a variety of heuristics
    and preferences that are specific to our scenario.

    Args:
        query (str): The common or scientific name we want to look up
        taxonomy_preference (str, optional): 'inat' or 'gbif'
        retry (bool, optional): if the initial lookup fails, should we try heuristic
            substitutions, e.g. replacing "_" with " ", or "spp" with "species"?

    Returns:
        TaxonomicMatch: the best taxonomic match, or None
    """

    m,query = _get_preferred_taxonomic_match(query=query,taxonomy_preference=taxonomy_preference)
    if (len(m.scientific_name) > 0) or (not retry):
        return m

    for s in hyphenated_terms:
        query = query.replace(' ' + s,'-' + s)
    m,query = _get_preferred_taxonomic_match(query=query,taxonomy_preference=taxonomy_preference)

    if (len(m.scientific_name) > 0) or (not retry):
        return m

    query = query.replace(' species','')
    query = query.replace(' order','')
    query = query.replace(' genus','')
    query = query.replace(' family','')
    query = query.replace(' subfamily','')
    m,query = _get_preferred_taxonomic_match(query=query,taxonomy_preference=taxonomy_preference)

    return m


def validate_and_convert(data):
    """
    Recursively validates that all elements in the nested structure are only
    tuples, lists, ints, or np.int64, and converts np.int64 to int.

    Args:
        data: The nested structure to validate and convert

    Returns:
        The validated and converted structure

    Raises:
        TypeError: If an invalid type is encountered
    """

    if isinstance(data, np.int64):
        return int(data)
    elif isinstance(data, int) or isinstance(data, str):
        return data
    elif isinstance(data, (list, tuple)):
        # Process lists and tuples recursively
        container_type = type(data)
        return container_type(validate_and_convert(item) for item in data)
    else:
        raise TypeError(f"Invalid type encountered: {type(data).__name__}. "
                        f"Only int, np.int64, list, and tuple are allowed.")

# ...def validate_and_convert(...)


def _get_preferred_taxonomic_match(query: str, taxonomy_preference = 'inat') -> TaxonomicMatch:

    query = query.lower().strip().replace('_', ' ')
    query = query.replace('unidentified','')
    query = query.replace('unknown','')
    if query.endswith(' sp'):
        query = query.replace(' sp','')
    if query.endswith(' group'):
        query = query.replace(' group','')

    query = query.strip()

    # query = 'person'
    matches = get_taxonomic_info(query)

    # Do we have an iNat match?
    inat_matches = [m for m in matches if m['source'] == 'inat']
    gbif_matches = [m for m in matches if m['source'] == 'gbif']

    # print_taxonomy_matches(inat_matches, verbose=True)
    # print_taxonomy_matches(gbif_matches, verbose=True)

    scientific_name = ''
    common_name = ''
    taxonomic_level = ''
    match = ''
    source = ''
    taxonomy_string = ''

    n_inat_matches = len(inat_matches)
    n_gbif_matches = len(gbif_matches)

    selected_matches = None

    assert taxonomy_preference in ['gbif','inat'],\
        'Unrecognized taxonomy preference: {}'.format(taxonomy_preference)

    if n_inat_matches > 0 and taxonomy_preference == 'inat':
        selected_matches = 'inat'
    elif n_gbif_matches > 0:
        selected_matches = 'gbif'

    if selected_matches == 'inat':

        i_match = 0

        if len(inat_matches) > 1:
            # print('Warning: multiple iNat matches for {}'.format(query))

            # Prefer chordates... most of the names that aren't what we want
            # are esoteric insects, like a moth called "cheetah"
            #
            # If we can't find a chordate, just take the first match.
            #
            # i_test_match = 0
            for i_test_match, match in enumerate(inat_matches):
                found_vertebrate = False
                taxonomy = match['taxonomy']
                for taxonomy_level in taxonomy:
                    taxon_rank = taxonomy_level[1]
                    scientific_name = taxonomy_level[2]
                    if taxon_rank == 'phylum' and scientific_name == 'chordata':
                        i_match = i_test_match
                        found_vertebrate = True
                        break
                if found_vertebrate:
                    break

        match = inat_matches[i_match]['taxonomy']

        # This is (taxonID, taxonLevel, scientific, [list of common])
        lowest_level = match[0]
        taxonomic_level = lowest_level[1]
        scientific_name = lowest_level[2]
        assert len(scientific_name) > 0
        common_names = lowest_level[3]
        if len(common_names) > 1:
            # print(f'Warning: multiple iNat common names for {query}')
            # Default to returning the query
            if query in common_names:
                common_name = query
            else:
                common_name = common_names[0]
        elif len(common_names) > 0:
            common_name = common_names[0]

        # print(f'Matched iNat {query} to {scientific_name},{common_name}')
        source = 'inat'

    # ...if we had iNat matches

    # If we either prefer GBIF or didn't have iNat matches
    #
    # Code is deliberately redundant here; I'm expecting some subtleties in how
    # handle GBIF and iNat.
    elif selected_matches == 'gbif':

        i_match = 0

        if len(gbif_matches) > 1:
            # print('Warning: multiple GBIF matches for {}'.format(query))

            # Prefer chordates... most of the names that aren't what we want
            # are esoteric insects, like a moth called "cheetah"
            #
            # If we can't find a chordate, just take the first match.
            #
            # i_test_match = 0
            for i_test_match, match in enumerate(gbif_matches):
                found_vertebrate = False
                taxonomy = match['taxonomy']
                for taxonomy_level in taxonomy:
                    taxon_rank = taxonomy_level[1]
                    scientific_name = taxonomy_level[2]
                    if taxon_rank == 'phylum' and scientific_name == 'chordata':
                        i_match = i_test_match
                        found_vertebrate = True
                        break
                if found_vertebrate:
                    break

        match = gbif_matches[i_match]['taxonomy']

        # This is (taxonID, taxonLevel, scientific, [list of common])
        lowest_level = match[0]
        taxonomic_level = lowest_level[1]
        scientific_name = lowest_level[2]
        assert len(scientific_name) > 0

        common_names = lowest_level[3]
        if len(common_names) > 1:
            # print(f'Warning: multiple GBIF common names for {query}')
            # Default to returning the query
            if query in common_names:
                common_name = query
            else:
                common_name = common_names[0]
        elif len(common_names) > 0:
            common_name = common_names[0]

        source = 'gbif'

    # ...if we needed to look in the GBIF taxonomy

    # Convert np.int64's to ints
    if match is not None:
        match = validate_and_convert(match)

    taxonomy_string = str(match)

    m = TaxonomicMatch(scientific_name, common_name, taxonomic_level, source,
                        taxonomy_string, match)

    if (m.taxonomic_level == 'subgenus' and \
        match[1][1] == 'genus' and \
        match[1][2] == m.scientific_name):
        print('Removing redundant subgenus {}'.format(scientific_name))
        m = pop_levels(m,1)

    return m,query

# ...def _get_preferred_taxonomic_match()


#%% Interactive drivers and debug

if False:

    #%% Initialization

    initialize_taxonomy_lookup()


    #%% Taxonomic lookup

    # query = 'lion'
    query = 'xenoperdix'
    matches = get_taxonomic_info(query)
    # print(matches)

    print_taxonomy_matches(matches,verbose=True)

    print('\n\n')

    # Print the taxonomy in the taxonomy spreadsheet format
    assert matches[1]['source'] == 'inat'
    t = str(matches[1]['taxonomy'])
    print(t)
    import clipboard; clipboard.copy(t)


    #%% Directly access the taxonomy tables

    taxon_ids = gbif_vernacular_to_taxon_id['lion']
    for taxon_id in taxon_ids:
        i_row = gbif_taxon_id_to_row[taxon_id]
        print(taxonomy_row_to_string(gbif_taxonomy.iloc[i_row]))


#%% Command-line driver

def main(): # noqa

    # Read command line inputs (absolute path)
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    input_file = args.input_file

    initialize_taxonomy_lookup()

    # Read the tokens from the input text file
    with open(input_file, 'r') as f:
        tokens = f.readlines()

    # Loop through each token and get scientific name
    for token in tokens:
        token = token.strip().lower()
        matches = get_taxonomic_info(token)
        print_taxonomy_matches(matches)

if __name__ == '__main__':
    main()



"""

ct_utils.py

Numeric/geometry/array utility functions.

"""

#%% Imports and constants

import inspect
import json
import math
import os
import builtins
import datetime
import tempfile
import shutil
import platform
import sys
import uuid

import jsonpickle
import numpy as np

from operator import itemgetter

# List of file extensions we'll consider images; comparisons will be case-insensitive
# (i.e., no need to include both .jpg and .JPG on this list).
image_extensions = ['.jpg', '.jpeg', '.gif', '.png']


#%% Functions

def truncate_float_array(xs, precision=3):
    """
    Truncates the fractional portion of each floating-point value in the array [xs]
    to a specific number of floating-point digits.

    Args:
        xs (list): list of floats to truncate
        precision (int, optional): the number of significant digits to preserve, should be >= 1

    Returns:
        list: list of truncated floats
    """

    return [truncate_float(x, precision=precision) for x in xs]


def round_float_array(xs, precision=3):
    """
    Truncates the fractional portion of each floating-point value in the array [xs]
    to a specific number of floating-point digits.

    Args:
        xs (list): list of floats to round
        precision (int, optional): the number of significant digits to preserve, should be >= 1

    Returns:
        list: list of rounded floats
    """

    return [round_float(x,precision) for x in xs]


def round_float(x, precision=3):
    """
    Convenience wrapper for the native Python round()

    Args:
        x (float): number to truncate
        precision (int, optional): the number of significant digits to preserve, should be >= 1

    Returns:
        float: rounded value
    """

    return round(x,precision)


def truncate_float(x, precision=3):
    """
    Truncates the fractional portion of a floating-point value to a specific number of
    floating-point digits.

    For example:

        truncate_float(0.0003214884) --> 0.000321
        truncate_float(1.0003214884) --> 1.000321

    This function is primarily used to achieve a certain float representation
    before exporting to JSON.

    Args:
        x (float): scalar to truncate
        precision (int, optional): the number of significant digits to preserve, should be >= 1

    Returns:
        float: truncated version of [x]
    """

    return math.floor(x * (10 ** precision)) / (10 ** precision)


def args_to_object(args, obj):
    """
    Copies all fields from a Namespace (typically the output from parse_args) to an
    object. Skips fields starting with _. Does not check existence in the target
    object.

    Args:
        args (argparse.Namespace): the namespace to convert to an object
        obj (object): object whose whose attributes will be updated

    Returns:
        object: the modified object (modified in place, but also returned)
    """

    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)

    return obj


def dict_to_object(d, obj):
    """
    Copies all fields from a dict to an object. Skips fields starting with _.
    Does not check existence in the target object.

    Args:
        d (dict): the dict to convert to an object
        obj (object): object whose whose attributes will be updated

    Returns:
        object: the modified object (modified in place, but also returned)
    """

    for k in d.keys():
        if not k.startswith('_'):
            setattr(obj, k, d[k])

    return obj


def pretty_print_object(obj, b_print=True):
    """
    Converts an arbitrary object to .json, optionally printing the .json representation.

    Args:
        obj (object): object to print
        b_print (bool, optional): whether to print the object

    Returns:
        str: .json reprepresentation of [obj]
    """

    # _ = pretty_print_object(obj)

    # TODO: it's sloppy that I'm making a module-wide change here, consider at least
    # recording these operations and re-setting them at the end of this function.
    jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
    a = jsonpickle.encode(obj)
    s = '{}'.format(a)
    if b_print:
        print(s)
    return s


def is_list_sorted(L, reverse=False): # noqa
    """
    Returns True if the list L appears to be sorted, otherwise False.

    Calling is_list_sorted(L,reverse=True) is the same as calling
    is_list_sorted(L.reverse(),reverse=False).

    Args:
        L (list): list to evaluate
        reverse (bool, optional): whether to reverse the list before evaluating sort status

    Returns:
        bool: True if the list L appears to be sorted, otherwise False
    """

    if reverse:
        return all(L[i] >= L[i + 1] for i in range(len(L)-1))
    else:
        return all(L[i] <= L[i + 1] for i in range(len(L)-1))


def json_serialize_datetime(obj):
    """
    Serializes datetime.datetime and datetime.date objects to ISO format.

    Args:
        obj (object): The object to serialize.

    Returns:
        str: The ISO format string representation of the datetime object.

    Raises:
        TypeError: If the object is not a datetime.datetime or datetime.date instance.
    """
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable by json_serialize_datetime")


def write_json(path,
               content,
               indent=1,
               force_str=False,
               serialize_datetimes=False,
               ensure_ascii=True,
               encoding='utf-8'):
    """
    Standardized wrapper for json.dump().

    Args:
        path (str): filename to write to
        content (object): object to dump
        indent (int, optional): indentation depth passed to json.dump
        force_str (bool, optional): whether to force string conversion for non-serializable objects
        serialize_datetimes (bool, optional): whether to serialize datetime objects to ISO format
        ensure_ascii (bool, optional): whether to ensure ASCII characters in the output
        encoding (str, optional): string encoding to use
    """

    default_handler = None

    if serialize_datetimes:
        default_handler = json_serialize_datetime
        if force_str:
            def serialize_or_str(obj):
                try:
                    return json_serialize_datetime(obj)
                except TypeError:
                    return str(obj)
            default_handler = serialize_or_str
    elif force_str:
        default_handler = str

    # Create the parent directory if necessary
    parent_dir = os.path.dirname(path)
    if len(parent_dir) > 0:
        os.makedirs(parent_dir, exist_ok=True)

    with open(path, 'w', newline='\n', encoding=encoding) as f:
        json.dump(content, f, indent=indent, default=default_handler, ensure_ascii=ensure_ascii)

# ...def write_json(...)


def convert_yolo_to_xywh(yolo_box):
    """
    Converts a YOLO format bounding box [x_center, y_center, w, h] to
    [x_min, y_min, width_of_box, height_of_box].

    Args:
        yolo_box (list): bounding box of format [x_center, y_center, width_of_box, height_of_box]

    Returns:
        list: bbox with coordinates represented as [x_min, y_min, width_of_box, height_of_box]
    """

    x_center, y_center, width_of_box, height_of_box = yolo_box
    x_min = x_center - width_of_box / 2.0
    y_min = y_center - height_of_box / 2.0
    return [x_min, y_min, width_of_box, height_of_box]


def convert_xywh_to_xyxy(api_box):
    """
    Converts an xywh bounding box (the MD output format) to an xyxy bounding box (the format
    produced by TF-based MD models).

    Args:
        api_box (list): bbox formatted as [x_min, y_min, width_of_box, height_of_box]

    Returns:
        list: bbox formatted as [x_min, y_min, x_max, y_max]
    """

    x_min, y_min, width_of_box, height_of_box = api_box
    x_max = x_min + width_of_box
    y_max = y_min + height_of_box
    return [x_min, y_min, x_max, y_max]


def get_iou(bb1, bb2):
    """
    Calculates the intersection over union (IoU) of two bounding boxes.

    Adapted from:

    https://stackoverflow.com/questions/25349178/calculating-percentage-of-bounding-box-overlap-for-image-detector-evaluation

    Args:
        bb1 (list): [x_min, y_min, width_of_box, height_of_box]
        bb2 (list): [x_min, y_min, width_of_box, height_of_box]

    Returns:
        float: intersection_over_union, a float in [0, 1]
    """

    bb1 = convert_xywh_to_xyxy(bb1)
    bb2 = convert_xywh_to_xyxy(bb2)

    assert bb1[0] < bb1[2], 'Malformed bounding box (x2 >= x1)'
    assert bb1[1] < bb1[3], 'Malformed bounding box (y2 >= y1)'

    assert bb2[0] < bb2[2], 'Malformed bounding box (x2 >= x1)'
    assert bb2[1] < bb2[3], 'Malformed bounding box (y2 >= y1)'

    # Determine the coordinates of the intersection rectangle
    x_left = max(bb1[0], bb2[0])
    y_top = max(bb1[1], bb2[1])
    x_right = min(bb1[2], bb2[2])
    y_bottom = min(bb1[3], bb2[3])

    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # Compute the area of both AABBs
    bb1_area = (bb1[2] - bb1[0]) * (bb1[3] - bb1[1])
    bb2_area = (bb2[2] - bb2[0]) * (bb2[3] - bb2[1])

    # Compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the intersection area.
    iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
    assert iou >= 0.0, 'Illegal IOU < 0'
    assert iou <= 1.0, 'Illegal IOU > 1'
    return iou


def _get_max_conf_from_detections(detections):
    """
    Internal function used by get_max_conf(); don't call this directly.
    """

    max_conf = 0.0
    if detections is not None and len(detections) > 0:
        confidences = [det['conf'] for det in detections]
        max_conf = max(confidences)
    return max_conf


def get_max_conf(im):
    """
    Given an image dict in the MD output format, computes the maximum detection confidence for any
    class.  Returns 0.0 if there were no detections, if there was a failure, or if 'detections' isn't
    present.

    Args:
        im (dict): image dictionary in the MD output format (with a 'detections' field)

    Returns:
        float: the maximum detection confidence across all classes
    """

    max_conf = 0.0
    if 'detections' in im and im['detections'] is not None and len(im['detections']) > 0:
        max_conf = _get_max_conf_from_detections(im['detections'])
    return max_conf


def sort_results_for_image(im):
    """
    Sort classification and detection results in descending order by confidence (in place).

    Args:
        im (dict): image dictionary in the MD output format (with a 'detections' field)
    """
    if 'detections' not in im or im['detections'] is None:
        return

    # Sort detections in descending order by confidence
    im['detections'] = sort_list_of_dicts_by_key(im['detections'],k='conf',reverse=True)

    for det in im['detections']:

        # Sort classifications (which are (class,conf) tuples) in descending order by confidence
        if 'classifications' in det and \
            (det['classifications'] is not None) and \
            (len(det['classifications']) > 0):
            classifications = det['classifications']
            det['classifications'] = \
                sorted(classifications,key=itemgetter(1),reverse=True)


def point_dist(p1,p2):
    """
    Computes the distance between two points, represented as length-two tuples.

    Args:
        p1 (list or tuple): point, formatted as (x,y)
        p2 (list or tuple): point, formatted as (x,y)

    Returns:
        float: the Euclidean distance between p1 and p2
    """

    return math.sqrt( ((p1[0]-p2[0])**2) + ((p1[1]-p2[1])**2) )


def rect_distance(r1, r2, format='x0y0x1y1'):
    """
    Computes the minimum distance between two axis-aligned rectangles, each represented as
    (x0,y0,x1,y1) by default.

    Can also specify "format" as x0y0wh for MD-style bbox formatting (x0,y0,w,h).

    Args:
        r1 (list or tuple): rectangle, formatted as (x0,y0,x1,y1) or (x0,y0,xy,y1)
        r2 (list or tuple): rectangle, formatted as (x0,y0,x1,y1) or (x0,y0,xy,y1)
        format (str, optional): whether the boxes are formatted as 'x0y0x1y1' (default) or 'x0y0wh'

    Returns:
        float: the minimum distance between r1 and r2
    """

    assert format in ('x0y0x1y1','x0y0wh'), 'Illegal rectangle format {}'.format(format)

    if format == 'x0y0wh':
        # Convert to x0y0x1y1 without modifying the original rectangles
        r1 = [r1[0],r1[1],r1[0]+r1[2],r1[1]+r1[3]]
        r2 = [r2[0],r2[1],r2[0]+r2[2],r2[1]+r2[3]]

    # https://stackoverflow.com/a/26178015
    x1, y1, x1b, y1b = r1
    x2, y2, x2b, y2b = r2
    left = x2b < x1
    right = x1b < x2
    bottom = y2b < y1
    top = y1b < y2
    if top and left:
        return point_dist((x1, y1b), (x2b, y2))
    elif left and bottom:
        return point_dist((x1, y1), (x2b, y2b))
    elif bottom and right:
        return point_dist((x1b, y1), (x2, y2b))
    elif right and top:
        return point_dist((x1b, y1b), (x2, y2))
    elif left:
        return x1 - x2b
    elif right:
        return x2 - x1b
    elif bottom:
        return y1 - y2b
    elif top:
        return y2 - y1b
    else:
        return 0.0


def split_list_into_fixed_size_chunks(L,n): # noqa
    """
    Split the list or tuple L into chunks of size n (allowing at most one chunk with size
    less than N, i.e. len(L) does not have to be a multiple of n).

    Args:
        L (list): list to split into chunks
        n (int): preferred chunk size

    Returns:
        list: list of chunks, where each chunk is a list of length n or n-1
    """

    return [L[i * n:(i + 1) * n] for i in range((len(L) + n - 1) // n )]


def split_list_into_n_chunks(L, n, chunk_strategy='greedy'): # noqa
    """
    Splits the list or tuple L into n equally-sized chunks (some chunks may be one
    element smaller than others, i.e. len(L) does not have to be a multiple of n).

    chunk_strategy can be "greedy" (default, if there are k samples per chunk, the first
    k go into the first chunk) or "balanced" (alternate between chunks when pulling
    items from the list).

    Args:
        L (list): list to split into chunks
        n (int): number of chunks
        chunk_strategy (str, optional): "greedy" or "balanced"; see above

    Returns:
        list: list of chunks, each of which is a list
    """

    if chunk_strategy == 'greedy':
        k, m = divmod(len(L), n)
        return list(L[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))
    elif chunk_strategy == 'balanced':
        chunks = [ [] for _ in range(n) ]
        for i_item,item in enumerate(L):
            i_chunk = i_item % n
            chunks[i_chunk].append(item)
        return chunks
    else:
        raise ValueError('Invalid chunk strategy: {}'.format(chunk_strategy))


def sort_list_of_dicts_by_key(L, k, reverse=False, none_handling='smallest'): # noqa ("L" should be lowercase)
    """
    Sorts the list of dictionaries [L] by the key [k].

    Args:
        L (list): list of dictionaries to sort
        k (object, typically str): the sort key
        reverse (bool, optional): whether to sort in reverse (descending) order
        none_handling (str, optional): how to handle None values. Options:
            "smallest" - treat None as smaller than all other values (default)
            "largest" - treat None as larger than all other values
            "error" - raise error when None is compared with non-None

    Returns:
        list: sorted copy of [L]
    """

    if none_handling == 'error':
        return sorted(L, key=lambda d: d[k], reverse=reverse)
    elif none_handling == 'smallest':
        # None values treated as smaller than other values: use tuple (is_not_none, value)
        return sorted(L, key=lambda d: (d[k] is not None, d[k]), reverse=reverse)
    elif none_handling == "largest":
        # None values treated as larger than other values: use tuple (is_none, value)
        return sorted(L, key=lambda d: (d[k] is None, d[k]), reverse=reverse)
    else:
        raise ValueError('Invalid none_handling value: {}'.format(none_handling))


def sort_dictionary_by_key(d,reverse=False):
    """
    Sorts the dictionary [d] by key.

    Args:
        d (dict): dictionary to sort
        reverse (bool, optional): whether to sort in reverse (descending) order

    Returns:
        dict: sorted copy of [d]
    """

    d = dict(sorted(d.items(),reverse=reverse))
    return d


def sort_dictionary_by_value(d,sort_values=None,reverse=False):
    """
    Sorts the dictionary [d] by value.  If sort_values is None, uses d.values(),
    otherwise uses the dictionary sort_values as the sorting criterion.  Always
    returns a new standard dict, so if [d] is, for example, a defaultdict, the
    returned value is not.

    Args:
        d (dict): dictionary to sort
        sort_values (dict, optional): dictionary mapping keys in [d] to sort values (defaults
            to None, uses [d] itself for sorting)
        reverse (bool, optional): whether to sort in reverse (descending) order

    Returns:
        dict: sorted copy of [d]
    """

    if sort_values is None:
        d = {k: v for k, v in sorted(d.items(), key=lambda item: item[1], reverse=reverse)}
    else:
        d = {k: v for k, v in sorted(d.items(), key=lambda item: sort_values[item[0]], reverse=reverse)}
    return d


def invert_dictionary(d, verify_unique=False):
    """
    Creates a new dictionary that maps d.values() to d.keys()

    Args:
        d (dict): dictionary to invert
        verify_unique (bool, optional): error if values are not unique

    Returns:
        dict: inverted copy of [d]
    """

    if verify_unique:
        n_unique_values = len(set(d.values()))
        assert len(d) == n_unique_values, \
            'Only {} unique values ({} total)'.format(
                n_unique_values,len(d))
    return {v: k for k, v in d.items()}


def round_floats_in_nested_dict(obj,
                                decimal_places=5,
                                allow_iterator_conversion=False):
    """
    Recursively rounds all floating point values in a nested structure to the
    specified number of decimal places. Handles dictionaries, lists, tuples,
    sets, and other iterables. Modifies mutable objects in place by default.

    Args:
        obj (obj): The object to process (can be a dict, list, set, tuple, or primitive value)
        decimal_places (int, optional): Number of decimal places to round to
        allow_iterator_conversion (bool, optional): for iterator types, should we convert
            to lists?  Otherwise we error.

    Returns:
        The processed object (useful for recursive calls)
    """

    if isinstance(obj, dict):
        for key in obj:
            obj[key] = round_floats_in_nested_dict(obj[key], decimal_places=decimal_places,
                                                   allow_iterator_conversion=allow_iterator_conversion)
        return obj

    elif isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = round_floats_in_nested_dict(obj[i], decimal_places=decimal_places,
                                                 allow_iterator_conversion=allow_iterator_conversion)
        return obj

    elif isinstance(obj, tuple):
        # Tuples are immutable, so we create a new one
        return tuple(round_floats_in_nested_dict(item, decimal_places=decimal_places,
                                                 allow_iterator_conversion=allow_iterator_conversion) for item in obj)

    elif isinstance(obj, set):
        # Sets are mutable but we can't modify elements in-place
        # Convert to list, process, and convert back to set
        return set(round_floats_in_nested_dict(list(obj), decimal_places=decimal_places,
                                               allow_iterator_conversion=allow_iterator_conversion))

    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        # Handle other iterable types: convert to list, process, and convert back
        processed_list = [round_floats_in_nested_dict(item,
                                                      decimal_places=decimal_places,
                                                      allow_iterator_conversion=allow_iterator_conversion) \
                                                        for item in obj]

        # Try to recreate the original type, but fall back to list for iterators
        try:
            return type(obj)(processed_list)
        except (TypeError, ValueError):
            if allow_iterator_conversion:
                # For iterators and other types that can't be reconstructed, return a list
                return processed_list
            else:
                raise ValueError('Cannot process iterator types when allow_iterator_conversion is False')

    elif isinstance(obj, float):
        return round(obj, decimal_places)

    else:
        # For other types (int, str, bool, None, etc.), return as is
        return obj

# ...def round_floats_in_nested_dict(...)


def image_file_to_camera_folder(image_fn):
    r"""
    Removes common overflow folders (e.g. RECNX101, RECNX102) from paths, i.e. turn:

    a\b\c\RECNX101\image001.jpg

    ...into:

    a\b\c

    Returns the same thing as os.dirname() (i.e., just the folder name) if no overflow folders are
    present.

    Always converts backslashes to slashes.

    Args:
        image_fn (str): the image filename from which we should remove overflow folders

    Returns:
        str: a version of [image_fn] from which camera overflow folders have been removed
    """

    import re

    # 100RECNX is the overflow folder style for Reconyx cameras
    # 100EK113 is (for some reason) the overflow folder style for Bushnell cameras
    # 100_BTCF is the overflow folder style for Browning cameras
    # 100MEDIA is the overflow folder style used on a number of consumer-grade cameras
    patterns = [r'/\d+RECNX/',r'/\d+EK\d+/',r'/\d+_BTCF/',r'/\d+MEDIA/']

    image_fn = image_fn.replace('\\','/')
    for pat in patterns:
        image_fn = re.sub(pat,'/',image_fn)
    camera_folder = os.path.dirname(image_fn)

    return camera_folder


def is_iterable(x):
    """
    Uses duck typing to assess whether [x] is iterable (list, set, dict, etc.).

    Args:
        x (object): the object to test

    Returns:
        bool: True if [x] appears to be iterable, otherwise False
    """

    try:
        _ = iter(x)
    except Exception:
       return False
    return True


def is_empty(v,strip_strings=True):
    """
    A common definition of "empty" used throughout the repo, particularly when loading
    data from .csv files.  "empty" includes None, '', and NaN.

    Args:
        v (obj): the object to evaluate for emptiness
        strip_strings (bool, optional): if v is a string, should whitespace be
            considered empty?

    Returns:
        bool: True if [v] is None, '', or NaN, otherwise False
    """
    if v is None:
        return True
    if isinstance(v,str) and len(v) == 0:
        return True
    if isinstance(v,str) and strip_strings and len(v.strip()) == 0:
        return True
    if isinstance(v,float) and np.isnan(v):
        return True
    return False


def to_bool(v):
    """
    Convert an object to a bool with specific rules.

    Args:
        v (object): The object to convert

    Returns:
        bool or None:
        - For strings: True if 'true' (case-insensitive), False if 'false', recursively applied if int-like
        - For int/bytes: False if 0, True otherwise
        - For bool: returns the bool as-is
        - For other types: None
    """

    if isinstance(v, bool):
        return v

    if isinstance(v, str):

        try:
            v = int(v)
            return to_bool(v)
        except Exception:
            pass

        v = v.lower().strip()
        if v == 'true':
            return True
        elif v == 'false':
            return False
        else:
            return None

    if isinstance(v, (int, bytes)):
        return v != 0

    return None


def min_none(a,b):
    """
    Returns the minimum of a and b.  If both are None, returns None.  If one is None,
    returns the other.

    Args:
        a (numeric): the first value to compare
        b (numeric): the second value to compare

    Returns:
        numeric: the minimum of a and b, or None
    """
    if a is None and b is None:
        return None
    elif a is None:
        return b
    elif b is None:
        return a
    else:
        return min(a,b)


def max_none(a,b):
    """
    Returns the maximum of a and b.  If both are None, returns None.  If one is None,
    returns the other.

    Args:
        a (numeric): the first value to compare
        b (numeric): the second value to compare

    Returns:
        numeric: the maximum of a and b, or None
    """
    if a is None and b is None:
        return None
    elif a is None:
        return b
    elif b is None:
        return a
    else:
        return max(a,b)


def isnan(v):
    """
    Returns True if v is a nan-valued float, otherwise returns False.

    Args:
        v (obj): the object to evaluate for nan-ness

    Returns:
        bool: True if v is a nan-valued float, otherwise False
    """

    try:
        return np.isnan(v)
    except Exception:
        return False


def compare_values_nan_equal(v0,v1):
    """
    Utility function for comparing two values when we want to return True if both
    values are NaN.

    Args:
        v0 (object): the first value to compare
        v1 (object): the second value to compare

    Returns:
        bool: True if v0 == v1, or if both v0 and v1 are NaN
    """

    if isinstance(v0,float) and isinstance(v1,float) and np.isnan(v0) and np.isnan(v1):
        return True
    return v0 == v1


def sets_overlap(set1, set2):
    """
    Determines whether two sets overlap.

    Args:
        set1 (set): the first set to compare (converted to a set if it's not already)
        set2 (set): the second set to compare (converted to a set if it's not already)

    Returns:
        bool: True if any elements are shared between set1 and set2
    """

    return not set(set1).isdisjoint(set(set2))


def is_function_name(s,calling_namespace):
    """
    Determines whether [s] is a callable function in the global or local scope, or a
    built-in function.

    Args:
        s (str): the string to test for function-ness
        calling_namespace (dict): typically pass the output of locals()
    """

    assert isinstance(s,str), 'Input is not a string'

    return callable(globals().get(s)) or \
        callable(locals().get(s)) or \
        callable(calling_namespace.get(s)) or \
        callable(getattr(builtins, s, None))


# From https://gist.github.com/fralau/061a4f6c13251367ef1d9a9a99fb3e8d
def parse_kvp(s,kv_separator='='):
    """
    Parse a key/value pair, separated by [kv_separator].  Errors if s is not
    a valid key/value pair string.  Strips leading/trailing whitespace from
    the key and value.

    Args:
        s (str): the string to parse
        kv_separator (str, optional): the string separating keys from values.

    Returns:
        tuple: a 2-tuple formatted as (key,value)
    """

    items = s.split(kv_separator)
    assert len(items) > 1, 'Illegal key-value pair: {}'.format(s)
    key = items[0].strip()
    if len(items) > 1:
        value = kv_separator.join(items[1:]).strip()
    return (key, value)


def parse_kvp_list(items,kv_separator='=',d=None):
    """
    Parse a list key-value pairs into a dictionary.  If items is None or [],
    returns {}.

    Args:
        items (list): the list of KVPs to parse
        kv_separator (str, optional): the string separating keys from values.
        d (dict, optional): the initial dictionary, defaults to {}

    Returns:
        dict: a dict mapping keys to values
    """

    if d is None:
        d = {}

    if items is None or len(items) == 0:
        return d

    for item in items:
        key, value = parse_kvp(item,kv_separator=kv_separator)
        d[key] = value

    return d


def dict_to_kvp_list(d,
                     item_separator=' ',
                     kv_separator='=',
                     non_string_value_handling='error'):
    """
    Convert a string <--> string dict into a string containing list of list of
    key-value pairs.  I.e., converts {'a':'dog','b':'cat'} to 'a=dog b=cat'.  If
    d is None, returns None.  If d is empty, returns ''.

    Args:
        d (dict): the dictionary to convert, must contain only strings
        item_separator (str, optional): the delimiter between KV pairs
        kv_separator (str, optional): the separator betweena a key and its value
        non_string_value_handling (str, optional): what do do with non-string values,
            can be "omit", "error", or "convert"

    Returns:
        str: the string representation of [d]
    """

    if d is None:
        return None

    if len(d) == 0:
        return ''

    s = None
    for k in d.keys():
        assert isinstance(k,str), 'Input {} is not a str <--> str dict'.format(str(d))
        v = d[k]
        if not isinstance(v,str):
            if non_string_value_handling == 'error':
                raise ValueError('Input {} is not a str <--> str dict'.format(str(d)))
            elif non_string_value_handling == 'omit':
                continue
            elif non_string_value_handling == 'convert':
                v = str(v)
            else:
                raise ValueError('Unrecognized non_string_value_handling value: {}'.format(
                    non_string_value_handling))
        if s is None:
            s = ''
        else:
            s += item_separator
        s += k + kv_separator + v

    if s is None:
        s = ''

    return s


def parse_bool_string(s, strict=False):
    """
    Convert the strings "true" or "false" to boolean values.  Case-insensitive, discards
    leading and trailing whitespace.  If s is already a bool, returns s.

    Args:
        s (str or bool): the string to parse, or the bool to return
        strict (bool, optional): only allow "true" or "false", otherwise
            handles "1", "0", "yes", and "no".

    Returns:
        bool: the parsed value
    """

    if isinstance(s,bool):
        return s
    s = str(s).lower().strip()

    if strict:
        # Fun fact: ('false') (rather than ('false,')) creates a string,
        # not a tuple.
        false_strings = ('false',)
        true_strings = ('true',)
    else:
        false_strings = ('no', 'false', 'f', 'n', '0')
        true_strings = ('yes', 'true', 't', 'y', '1')
    if s in true_strings:
        return True
    elif s in false_strings:
        return False
    else:
        raise ValueError('Cannot parse bool from string {}'.format(str(s)))


def make_temp_folder(top_level_folder='megadetector',subfolder=None,append_guid=True):
    """
    Creates a temporary folder within the system temp folder, by default in a subfolder
    called megadetector/some_guid.  Used for testing without making too much of a mess.

    Args:
        top_level_folder (str, optional): the top-level folder to use within the system temp folder
        subfolder (str, optional): the subfolder within [top_level_folder]
        append_guid (bool, optional): append a guid to the subfolder

    Returns:
        str: the new directory
    """

    to_return = os.path.join(tempfile.gettempdir(),top_level_folder)
    if subfolder is not None:
        to_return = os.path.join(to_return,subfolder)
    if append_guid:
        to_return = os.path.join(to_return,str(uuid.uuid1()))
    to_return = os.path.normpath(to_return)
    os.makedirs(to_return,exist_ok=True)
    return to_return


def make_test_folder(subfolder=None):
    """
    Wrapper around make_temp_folder that creates folders within megadetector/tests

    Args:
        subfolder (str): specific subfolder to create within the default megadetector temp
            folder.
    """

    return make_temp_folder(top_level_folder='megadetector/tests',
                            subfolder=subfolder,
                            append_guid=True)


#%% Environment utilities

def is_sphinx_build():
    """
    Determine whether we are running in the context of our Sphinx build.

    Returns:
        bool: True if we're running a Sphinx build
    """

    is_sphinx = hasattr(builtins, '__sphinx_build__')
    return is_sphinx


def is_running_in_gha():
    """
    Determine whether we are running on a GitHub Actions runner.

    Returns:
        bool: True if we're running in a GHA runner
    """

    running_in_gha = False

    if ('GITHUB_ACTIONS' in os.environ):
        # Documentation is inconsistent on how this variable presents itself
        if isinstance(os.environ['GITHUB_ACTIONS'],bool) and \
            os.environ['GITHUB_ACTIONS']:
            running_in_gha = True
        elif isinstance(os.environ['GITHUB_ACTIONS'],str) and \
            os.environ['GITHUB_ACTIONS'].lower() == ('true'):
            running_in_gha = True

    return running_in_gha


def environment_is_wsl():
    """
    Determines whether we're running in WSL.

    Returns:
        True if we're running in WSL
    """

    if sys.platform not in ('linux','posix'):
        return False
    platform_string = ' '.join(platform.uname()).lower()
    return 'microsoft' in platform_string and 'wsl' in platform_string



#%% Tests

def test_write_json():
    """
    Test driver for write_json.
    """

    temp_dir = make_test_folder()

    def _verify_json_file(file_path, expected_content_str):
        with open(file_path, 'r', encoding='utf-8') as f:
            content = json.load(f)
        assert isinstance(content,dict)
        content = sort_dictionary_by_key(content)
        expected_content = json.loads(expected_content_str)
        expected_content = sort_dictionary_by_key(expected_content)
        assert content == expected_content, \
            f"File {file_path} content mismatch.\nExpected:\n{expected_content}\nGot:\n{content}"

    # Test default indent (1)
    data_default = {'a': 1, 'b': 2}
    file_path_default = os.path.join(temp_dir, 'test_default_indent.json')
    write_json(file_path_default, data_default)
    # Default indent is 1
    _verify_json_file(file_path_default, '{\n "a": 1,\n "b": 2\n}')

    # Test custom indent (e.g., 4)
    data_custom_indent = {'a': 1, 'b': 2}
    file_path_custom_indent = os.path.join(temp_dir, 'test_custom_indent.json')
    write_json(file_path_custom_indent, data_custom_indent, indent=4)
    _verify_json_file(file_path_custom_indent, '{\n    "a": 1,\n    "b": 2\n}')

    # Test indent=None (compact)
    data_no_indent = {'a': 1, 'b': 2}
    file_path_no_indent = os.path.join(temp_dir, 'test_no_indent.json')
    write_json(file_path_no_indent, data_no_indent, indent=None)
    _verify_json_file(file_path_no_indent, '{"a": 1, "b": 2}')

    # Test force_str=True
    data_force_str = {'a': 1, 's': {1, 2, 3}}  # Set is not normally JSON serializable
    file_path_force_str = os.path.join(temp_dir, 'test_force_str.json')
    write_json(file_path_force_str, data_force_str, force_str=True)
    with open(file_path_force_str, 'r', encoding='utf-8') as f:
        result_force_str = json.load(f)
    assert isinstance(result_force_str['s'], str)
    assert eval(result_force_str['s']) == {1, 2, 3}

    # Test serialize_datetimes=True
    dt = datetime.datetime(2023, 1, 1, 10, 30, 0)
    d_date = datetime.date(2023, 2, 15)
    data_serialize_datetimes = {'dt_obj': dt, 'd_obj': d_date}
    file_path_serialize_datetimes = os.path.join(temp_dir, 'test_serialize_datetimes.json')
    write_json(file_path_serialize_datetimes, data_serialize_datetimes, serialize_datetimes=True)
    _verify_json_file(file_path_serialize_datetimes, '{\n "d_obj": "2023-02-15",\n "dt_obj": "2023-01-01T10:30:00"\n}')

    # Test serialize_datetimes=True and force_str=True
    dt_combo = datetime.datetime(2023, 1, 1, 12, 0, 0)
    data_datetime_force_str = {'dt_obj': dt_combo, 's_obj': {4, 5}}
    file_path_datetime_force_str = os.path.join(temp_dir, 'test_datetime_and_force_str.json')
    write_json(file_path_datetime_force_str, data_datetime_force_str, serialize_datetimes=True, force_str=True)
    with open(file_path_datetime_force_str, 'r', encoding='utf-8') as f:
        result_datetime_force_str = json.load(f)
    assert result_datetime_force_str['dt_obj'] == "2023-01-01T12:00:00"
    assert isinstance(result_datetime_force_str['s_obj'], str)
    assert eval(result_datetime_force_str['s_obj']) == {4, 5}

    # Test ensure_ascii=False (with non-ASCII chars)
    data_ensure_ascii_false = {'name': 'Jules César'}
    file_path_ensure_ascii_false = os.path.join(temp_dir, 'test_ensure_ascii_false.json')
    write_json(file_path_ensure_ascii_false, data_ensure_ascii_false, ensure_ascii=False)
    with open(file_path_ensure_ascii_false, 'r', encoding='utf-8') as f:
        content_ensure_ascii_false = f.read()
    assert content_ensure_ascii_false == '{\n "name": "Jules César"\n}'

    # Test ensure_ascii=True (with non-ASCII chars, default)
    data_ensure_ascii_true = {'name': 'Jules César'}
    file_path_ensure_ascii_true = os.path.join(temp_dir, 'test_ensure_ascii_true.json')
    write_json(file_path_ensure_ascii_true, data_ensure_ascii_true, ensure_ascii=True)
    with open(file_path_ensure_ascii_true, 'r', encoding='utf-8') as f:
        content_ensure_ascii_true = f.read()
    assert content_ensure_ascii_true == '{\n "name": "Jules C\\u00e9sar"\n}'

    shutil.rmtree(temp_dir)

# ...def test_write_json(...)


def test_path_operations():
    """
    Test path manipulation functions.
    """

    ##%% Camera folder mapping
    assert image_file_to_camera_folder('a/b/c/d/100EK113/blah.jpg') == 'a/b/c/d'
    assert image_file_to_camera_folder('a/b/c/d/100RECNX/blah.jpg') == 'a/b/c/d'
    assert image_file_to_camera_folder('a/b/c/d/blah.jpg') == 'a/b/c/d'
    assert image_file_to_camera_folder(r'a\b\c\d\100RECNX\blah.jpg') == 'a/b/c/d'


def test_geometric_operations():
    """
    Test geometric calculations like distances.
    """

    ##%% Test a few rectangle distances

    r1 = [0,0,1,1]; r2 = [0,0,1,1]; assert rect_distance(r1,r2)==0
    r1 = [0,0,1,1]; r2 = [0,0,1,100]; assert rect_distance(r1,r2)==0
    r1 = [0,0,1,1]; r2 = [1,1,2,2]; assert rect_distance(r1,r2)==0
    r1 = [0,0,1,1]; r2 = [1.1,0,0,1.1]; assert abs(rect_distance(r1,r2)-.1) < 0.00001

    r1 = [0.4,0.8,10,22]; r2 = [100, 101, 200, 210.4]; assert abs(rect_distance(r1,r2)-119.753) < 0.001
    r1 = [0.4,0.8,10,22]; r2 = [101, 101, 200, 210.4]; assert abs(rect_distance(r1,r2)-120.507) < 0.001
    r1 = [0.4,0.8,10,22]; r2 = [120, 120, 200, 210.4]; assert abs(rect_distance(r1,r2)-147.323) < 0.001

    # Test with 'x0y0wh' format
    r1_wh = [0,0,1,1]; r2_wh = [1,0,1,1]; assert rect_distance(r1_wh, r2_wh, format='x0y0wh') == 0
    r1_wh = [0,0,1,1]; r2_wh = [1.5,0,1,1]; assert abs(rect_distance(r1_wh, r2_wh, format='x0y0wh') - 0.5) < 0.00001


    ##%% Test point_dist

    assert point_dist((0,0), (3,4)) == 5.0
    assert point_dist((1,1), (1,1)) == 0.0


def test_dictionary_operations():
    """
    Test dictionary manipulation and sorting functions.
    """

    ##%% Test sort_list_of_dicts_by_key

    x = [{'a':5},{'a':0},{'a':10}]
    k = 'a'
    sorted_x = sort_list_of_dicts_by_key(x, k)
    assert sorted_x[0]['a'] == 0; assert sorted_x[1]['a'] == 5; assert sorted_x[2]['a'] == 10
    sorted_x_rev = sort_list_of_dicts_by_key(x, k, reverse=True)
    assert sorted_x_rev[0]['a'] == 10; assert sorted_x_rev[1]['a'] == 5; assert sorted_x_rev[2]['a'] == 0


    ##%% Test sort_dictionary_by_key

    d_key = {'b': 2, 'a': 1, 'c': 3}
    sorted_d_key = sort_dictionary_by_key(d_key)
    assert list(sorted_d_key.keys()) == ['a', 'b', 'c']
    sorted_d_key_rev = sort_dictionary_by_key(d_key, reverse=True)
    assert list(sorted_d_key_rev.keys()) == ['c', 'b', 'a']


    ##%% Test sort_dictionary_by_value

    d_val = {'a': 2, 'b': 1, 'c': 3}
    sorted_d_val = sort_dictionary_by_value(d_val)
    assert list(sorted_d_val.keys()) == ['b', 'a', 'c']
    sorted_d_val_rev = sort_dictionary_by_value(d_val, reverse=True)
    assert list(sorted_d_val_rev.keys()) == ['c', 'a', 'b']

    # With sort_values
    sort_vals = {'a': 10, 'b': 0, 'c': 5}
    sorted_d_custom = sort_dictionary_by_value(d_val, sort_values=sort_vals)
    assert list(sorted_d_custom.keys()) == ['b', 'c', 'a']


    ##%% Test invert_dictionary

    d_inv = {'a': 'x', 'b': 'y'}
    inverted_d = invert_dictionary(d_inv)
    assert inverted_d == {'x': 'a', 'y': 'b'}

    # Does not check for uniqueness, last one wins
    d_inv_dup = {'a': 'x', 'b': 'x'}
    inverted_d_dup = invert_dictionary(d_inv_dup)
    assert inverted_d_dup == {'x': 'b'}


def test_float_rounding_and_truncation():
    """
    Test float rounding, truncation, and nested rounding functions.
    """

    ##%% Test round_floats_in_nested_dict

    data = {
        "name": "Project X",
        "values": [1.23456789, 2.3456789],
        "tuple_values": (3.45678901, 4.56789012),
        "set_values": {5.67890123, 6.78901234}, # Order not guaranteed in set, test min/max
        "metrics": {
            "score": 98.7654321,
            "components": [5.6789012, 6.7890123]
        },
        "other_iter": iter([7.89012345]) # Test other iterables
    }

    result = round_floats_in_nested_dict(data, decimal_places=5, allow_iterator_conversion=True)
    assert result['values'][0] == 1.23457
    assert result['tuple_values'][0] == 3.45679

    # For sets, convert to list and sort for consistent testing
    assert sorted(list(result['set_values'])) == sorted([5.67890, 6.78901])
    assert result['metrics']['score'] == 98.76543

    # Test other iterables by converting back to list
    assert list(result['other_iter'])[0] == 7.89012


    ##%% Test truncate_float_array and truncate_float

    assert truncate_float_array([0.12345, 0.67890], precision=3) == [0.123, 0.678]
    assert truncate_float_array([1.0, 2.0], precision=2) == [1.0, 2.0]
    assert truncate_float(0.12345, precision=3) == 0.123
    assert truncate_float(1.999, precision=2) == 1.99
    assert truncate_float(0.0003214884, precision=6) == 0.000321
    assert truncate_float(1.0003214884, precision=6) == 1.000321


    ##%% Test round_float_array and round_float

    assert round_float_array([0.12345, 0.67890], precision=3) == [0.123, 0.679]
    assert round_float_array([1.0, 2.0], precision=2) == [1.0, 2.0]
    assert round_float(0.12345, precision=3) == 0.123
    assert round_float(0.12378, precision=3) == 0.124
    assert round_float(1.999, precision=2) == 2.00


def test_object_conversion_and_presentation():
    """
    Test functions that convert or present objects.
    """

    ##%% Test args_to_object

    class ArgsObject:
        pass
    args_namespace = type('ArgsNameSpace', (), {'a': 1, 'b': 'test', '_c': 'ignored'})
    obj = ArgsObject()
    args_to_object(args_namespace, obj)
    assert obj.a == 1
    assert obj.b == 'test'
    assert not hasattr(obj, '_c')


    ##%% Test dict_to_object

    class DictObject:
        pass
    d = {'a': 1, 'b': 'test', '_c': 'ignored'}
    obj = DictObject()
    dict_to_object(d, obj)
    assert obj.a == 1
    assert obj.b == 'test'
    assert not hasattr(obj, '_c')


    ##%% Test pretty_print_object

    class PrettyPrintable:
        def __init__(self):
            self.a = 1
            self.b = "test"
    obj_to_print = PrettyPrintable()
    json_str = pretty_print_object(obj_to_print, b_print=False)

    # Basic check for valid json and presence of attributes
    parsed_json = json.loads(json_str) # Relies on json.loads
    assert parsed_json['a'] == 1
    assert parsed_json['b'] == "test"


def test_list_operations():
    """
    Test list sorting and chunking functions.
    """

    ##%% Test is_list_sorted

    assert is_list_sorted([1, 2, 3])
    assert not is_list_sorted([1, 3, 2])
    assert is_list_sorted([3, 2, 1], reverse=True)
    assert not is_list_sorted([1, 2, 3], reverse=True)
    assert is_list_sorted([]) # Empty list is considered sorted
    assert is_list_sorted([1]) # Single element list is sorted
    assert is_list_sorted([1,1,1])
    assert is_list_sorted([1,1,1], reverse=True)


    ##%% Test split_list_into_fixed_size_chunks

    assert split_list_into_fixed_size_chunks([1,2,3,4,5,6], 2) == [[1,2],[3,4],[5,6]]
    assert split_list_into_fixed_size_chunks([1,2,3,4,5], 2) == [[1,2],[3,4],[5]]
    assert split_list_into_fixed_size_chunks([], 3) == []
    assert split_list_into_fixed_size_chunks([1,2,3], 5) == [[1,2,3]]


    ##%% Test split_list_into_n_chunks

    # Greedy
    assert split_list_into_n_chunks([1,2,3,4,5,6], 3, chunk_strategy='greedy') == [[1,2],[3,4],[5,6]]
    assert split_list_into_n_chunks([1,2,3,4,5], 3, chunk_strategy='greedy') == [[1,2],[3,4],[5]]
    assert split_list_into_n_chunks([1,2,3,4,5,6,7], 3, chunk_strategy='greedy') == [[1,2,3],[4,5],[6,7]]
    assert split_list_into_n_chunks([], 3) == [[],[],[]]

    # Balanced
    assert split_list_into_n_chunks([1,2,3,4,5,6], 3, chunk_strategy='balanced') == [[1,4],[2,5],[3,6]]
    assert split_list_into_n_chunks([1,2,3,4,5], 3, chunk_strategy='balanced') == [[1,4],[2,5],[3]]
    assert split_list_into_n_chunks([], 3, chunk_strategy='balanced') == [[],[],[]]
    try:
        split_list_into_n_chunks([1,2,3], 2, chunk_strategy='invalid')
        raise AssertionError("ValueError not raised for invalid chunk_strategy")
    except ValueError:
        pass


def test_datetime_serialization():
    """
    Test datetime serialization functions.
    """

    ##%% Test json_serialize_datetime

    now = datetime.datetime.now()
    today = datetime.date.today()
    assert json_serialize_datetime(now) == now.isoformat()
    assert json_serialize_datetime(today) == today.isoformat()
    try:
        json_serialize_datetime("not a datetime")
        raise AssertionError("TypeError not raised for non-datetime object")
    except TypeError:
        pass
    try:
        json_serialize_datetime(123)
        raise AssertionError("TypeError not raised for non-datetime object")
    except TypeError:
        pass


def test_bounding_box_operations():
    """
    Test bounding box conversion and IoU calculation.
    """

    ##%% Test convert_yolo_to_xywh

    # [x_center, y_center, w, h]
    yolo_box = [0.5, 0.5, 0.2, 0.2]
    # [x_min, y_min, width_of_box, height_of_box]
    expected_xywh = [0.4, 0.4, 0.2, 0.2]
    assert np.allclose(convert_yolo_to_xywh(yolo_box), expected_xywh)


    ##%% Test convert_xywh_to_xyxy

    # [x_min, y_min, width_of_box, height_of_box]
    xywh_box = [0.1, 0.1, 0.3, 0.3]
    # [x_min, y_min, x_max, y_max]
    expected_xyxy = [0.1, 0.1, 0.4, 0.4]
    assert np.allclose(convert_xywh_to_xyxy(xywh_box), expected_xyxy)


    ##%% Test get_iou

    bb1 = [0, 0, 0.5, 0.5]  # x, y, w, h
    bb2 = [0.25, 0.25, 0.5, 0.5]
    assert abs(get_iou(bb1, bb2) - 0.142857) < 1e-5
    bb3 = [0, 0, 1, 1]
    bb4 = [0.5, 0.5, 1, 1]
    assert abs(get_iou(bb3, bb4) - (0.25 / 1.75)) < 1e-5
    bb5 = [0,0,1,1]
    bb6 = [1,1,1,1] # No overlap
    assert get_iou(bb5, bb6) == 0.0

    # Test malformed boxes (should ideally raise error or handle gracefully based on spec, current impl asserts)
    bb_malformed1 = [0.6, 0.0, 0.5, 0.5] # x_min > x_max after conversion
    bb_ok = [0.0, 0.0, 0.5, 0.5]
    try:
        get_iou(bb_malformed1, bb_ok)
        # This assert False will only be reached if the expected AssertionError is not raised by get_iou
        # assert False, "AssertionError for malformed bounding box (x2 >= x1) not raised in get_iou"
    except AssertionError as e:
        assert 'Malformed bounding box' in str(e)


def test_detection_processing():
    """
    Test functions related to processing detection results.
    """

    ##%% Test _get_max_conf_from_detections and get_max_conf

    detections1 = [{'conf': 0.8}, {'conf': 0.9}, {'conf': 0.75}]
    assert _get_max_conf_from_detections(detections1) == 0.9
    assert _get_max_conf_from_detections([]) == 0.0
    assert _get_max_conf_from_detections(None) == 0.0

    im1 = {'detections': detections1}
    assert get_max_conf(im1) == 0.9
    im2 = {'detections': []}
    assert get_max_conf(im2) == 0.0
    im3 = {} # No 'detections' key
    assert get_max_conf(im3) == 0.0
    im4 = {'detections': None}
    assert get_max_conf(im4) == 0.0


    ##%% Test sort_results_for_image

    img_data = {
        'detections': [
            {'conf': 0.7, 'classifications': [('c', 0.6), ('a', 0.9), ('b', 0.8)]},
            {'conf': 0.9, 'classifications': [('x', 0.95), ('y', 0.85)]},
            {'conf': 0.8} # No classifications field
        ]
    }
    sort_results_for_image(img_data)

    # Check detections sorted by conf
    assert img_data['detections'][0]['conf'] == 0.9
    assert img_data['detections'][1]['conf'] == 0.8
    assert img_data['detections'][2]['conf'] == 0.7

    # Check classifications sorted by conf (only for the first original detection, now at index 0 after sort)
    assert img_data['detections'][0]['classifications'][0] == ('x', 0.95)
    assert img_data['detections'][0]['classifications'][1] == ('y', 0.85)

    # Check classifications for the second original detection (now at index 2)
    assert img_data['detections'][2]['classifications'][0] == ('a', 0.9)
    assert img_data['detections'][2]['classifications'][1] == ('b', 0.8)
    assert img_data['detections'][2]['classifications'][2] == ('c', 0.6)

    # Test with no detections or no classifications field
    img_data_no_det = {'detections': None}
    sort_results_for_image(img_data_no_det)
    assert img_data_no_det['detections'] is None

    img_data_empty_det = {'detections': []}
    sort_results_for_image(img_data_empty_det)
    assert img_data_empty_det['detections'] == []

    img_data_no_classifications_field = {'detections': [{'conf': 0.8}]}
    sort_results_for_image(img_data_no_classifications_field)
    assert 'classifications' not in img_data_no_classifications_field['detections'][0]

    img_data_none_classifications = {'detections': [{'conf': 0.8, 'classifications':None}]}
    sort_results_for_image(img_data_none_classifications)
    assert img_data_none_classifications['detections'][0]['classifications'] is None

    img_data_empty_classifications = {'detections': [{'conf': 0.8, 'classifications':[]}]}
    sort_results_for_image(img_data_empty_classifications)
    assert img_data_empty_classifications['detections'][0]['classifications'] == []


def test_type_checking_and_validation():
    """
    Test type checking and validation utility functions.
    """

    ##%% Test is_iterable

    assert is_iterable([1,2,3])
    assert is_iterable("hello")
    assert is_iterable({'a':1})
    assert is_iterable(range(5))
    assert not is_iterable(123)
    assert not is_iterable(None)
    assert is_iterable(np.array([1,2]))


    ##%% Test is_empty

    assert is_empty(None)
    assert is_empty("")
    assert is_empty(" ", strip_strings=True)
    assert is_empty(np.nan)
    assert not is_empty(0)
    assert not is_empty(" ", strip_strings=False)
    assert not is_empty([])
    assert not is_empty({})
    assert not is_empty(False) # False is not empty


    ##%% Test min_none and max_none

    assert min_none(1, 2) == 1
    assert min_none(None, 2) == 2
    assert min_none(1, None) == 1
    assert min_none(None, None) is None
    assert max_none(1, 2) == 2
    assert max_none(None, 2) == 2
    assert max_none(1, None) == 1
    assert max_none(None, None) is None


    ##%% Test isnan

    assert isnan(np.nan)
    assert not isnan(0.0)
    assert not isnan("text")
    assert not isnan(None)
    assert not isnan(float('inf'))
    assert not isnan(float('-inf'))


    ##%% Test sets_overlap

    assert sets_overlap({1,2,3}, {3,4,5})
    assert not sets_overlap({1,2}, {3,4})
    assert sets_overlap([1,2,3], [3,4,5]) # Test with lists
    assert sets_overlap(set(), {1}) is False
    assert sets_overlap({1},{1})


    ##%% Test is_function_name

    def _test_local_func(): pass
    assert is_function_name("sets_overlap", locals()) # Test a function in ct_utils
    assert is_function_name("_test_local_func", locals()) # Test a local function
    assert is_function_name("print", locals()) # Test a builtin
    assert not is_function_name("non_existent_func", locals())
    global _test_global_func_ct_utils # Renamed to avoid conflict if run multiple times
    def _test_global_func_ct_utils(): pass
    assert is_function_name("_test_global_func_ct_utils", globals())
    # Clean up global
    del _test_global_func_ct_utils


def test_string_parsing():
    """
    Test string parsing utilities like KVP and boolean parsing.
    """

    ##%% Test parse_kvp and parse_kvp_list

    assert parse_kvp("key=value") == ("key", "value")
    assert parse_kvp("key = value with spaces") == ("key", "value with spaces")
    assert parse_kvp("key=value1=value2", kv_separator='=') == ("key", "value1=value2")
    try:
        parse_kvp("keyvalue")
        raise AssertionError("AssertionError not raised for invalid KVP")
    except AssertionError:
        pass

    kvp_list = ["a=1", "b = 2", "c=foo=bar"]
    parsed_list = parse_kvp_list(kvp_list)
    assert parsed_list == {"a": "1", "b": "2", "c": "foo=bar"}
    assert parse_kvp_list(None) == {}
    assert parse_kvp_list([]) == {}
    d_initial = {'z': '0'}

    # parse_kvp_list modifies d in place if provided
    parse_kvp_list(kvp_list, d=d_initial)
    assert d_initial == {"z": "0", "a": "1", "b": "2", "c": "foo=bar"}

    # Test with a different separator
    assert parse_kvp("key:value", kv_separator=":") == ("key", "value")
    assert parse_kvp_list(["a:1","b:2"], kv_separator=":") == {"a":"1", "b":"2"}


    ##%% Test dict_to_kvp_list

    d_kvp = {"a": "1", "b": "dog", "c": "foo=bar"}
    kvp_str = dict_to_kvp_list(d_kvp)

    # Order isn't guaranteed, so check for presence of all items and length
    assert "a=1" in kvp_str
    assert "b=dog" in kvp_str
    assert "c=foo=bar" in kvp_str
    assert len(kvp_str.split(' ')) == 3

    assert dict_to_kvp_list({}) == ""
    assert dict_to_kvp_list(None) is None
    d_kvp_int = {"a":1, "b":"text"}
    try:
        dict_to_kvp_list(d_kvp_int, non_string_value_handling='error')
        raise AssertionError("ValueError not raised for non-string value with 'error' handling")
    except ValueError:
        pass
    convert_result = dict_to_kvp_list(d_kvp_int, non_string_value_handling='convert')
    assert "a=1" in convert_result and "b=text" in convert_result

    omit_result = dict_to_kvp_list({"a":1, "b":"text"}, non_string_value_handling='omit')
    assert "a=1" not in omit_result and "b=text" in omit_result
    assert omit_result == "b=text"

    assert dict_to_kvp_list({"key":"val"}, item_separator="&", kv_separator=":") == "key:val"


    ##%% Test parse_bool_string

    assert parse_bool_string("true")
    assert parse_bool_string("True")
    assert parse_bool_string(" TRUE ")
    assert not parse_bool_string("false")
    assert not parse_bool_string("False")
    assert not parse_bool_string(" FALSE ")
    assert parse_bool_string("1", strict=False)
    assert not parse_bool_string("0", strict=False)
    assert parse_bool_string(True) is True # Test with existing bool
    assert parse_bool_string(False) is False
    try:
        parse_bool_string("maybe")
        raise AssertionError("ValueError not raised for invalid bool string")
    except ValueError:
        pass
    try:
        parse_bool_string("1",strict=True)
        raise AssertionError("ValueError not raised for '1'")
    except ValueError:
        pass


def test_temp_folder_creation():
    """
    Test temporary folder creation and cleanup.
    """

    # Store original tempdir for restoration if modified by tests (though unlikely for make_temp_folder)
    original_tempdir = tempfile.gettempdir()

    # Test make_temp_folder
    custom_top_level = "my_custom_temp_app_test" # Unique name for this test run
    custom_subfolder = "specific_test_run"

    # Test with default subfolder (UUID)
    temp_folder1_base = os.path.join(tempfile.gettempdir(), custom_top_level)
    temp_folder1 = make_temp_folder(top_level_folder=custom_top_level)
    assert os.path.exists(temp_folder1)
    assert os.path.basename(os.path.dirname(temp_folder1)) == custom_top_level
    assert temp_folder1_base == os.path.dirname(temp_folder1) # Path up to UUID should match

    # Cleanup: remove the custom_top_level which contains the UUID folder
    if os.path.exists(temp_folder1_base):
        shutil.rmtree(temp_folder1_base)
    assert not os.path.exists(temp_folder1_base)


    # Test with specified subfolder
    temp_folder2_base = os.path.join(tempfile.gettempdir(), custom_top_level)
    temp_folder2 = make_temp_folder(top_level_folder=custom_top_level,
                                    subfolder=custom_subfolder,
                                    append_guid=False)
    assert os.path.exists(temp_folder2)
    assert os.path.basename(temp_folder2) == custom_subfolder
    assert os.path.basename(os.path.dirname(temp_folder2)) == custom_top_level
    assert temp_folder2 == os.path.join(tempfile.gettempdir(), custom_top_level, custom_subfolder)

    # Cleanup
    if os.path.exists(temp_folder2_base):
        shutil.rmtree(temp_folder2_base)
    assert not os.path.exists(temp_folder2_base)


    # Test make_test_folder (which uses 'megadetector/tests' as top_level)
    #
    # This will create tempfile.gettempdir()/megadetector/tests/some_uuid or specified_subfolder
    megadetector_temp_base = os.path.join(tempfile.gettempdir(), "megadetector")
    test_subfolder = "my_specific_module_test"

    # Test with default subfolder for make_test_folder
    test_folder1 = make_test_folder() # Creates megadetector/tests/uuid_folder
    assert os.path.exists(test_folder1)
    assert os.path.basename(os.path.dirname(test_folder1)) == "tests"
    assert os.path.basename(os.path.dirname(os.path.dirname(test_folder1))) == "megadetector"

    # Cleanup for make_test_folder default: remove the 'megadetector' base temp dir
    if os.path.exists(megadetector_temp_base):
        shutil.rmtree(megadetector_temp_base)
    assert not os.path.exists(megadetector_temp_base)


    # Test with specified subfolder for make_test_folder
    test_folder2 = make_test_folder(subfolder=test_subfolder) # megadetector/tests/my_specific_module_test
    assert os.path.exists(test_folder2)
    assert test_subfolder in test_folder2
    assert "megadetector" in test_folder2

    # Cleanup for make_test_folder specific: remove the 'megadetector' base temp dir
    if os.path.exists(megadetector_temp_base):
        shutil.rmtree(megadetector_temp_base)
    assert not os.path.exists(megadetector_temp_base)

    # Verify cleanup if top level folder was 'megadetector' (default for make_temp_folder)
    #
    # This means it creates tempfile.gettempdir()/megadetector/uuid_folder
    default_temp_folder = make_temp_folder()
    assert os.path.exists(default_temp_folder)
    assert os.path.basename(os.path.dirname(default_temp_folder)) == "megadetector"

    # Cleanup: remove the 'megadetector' base temp dir created by default make_temp_folder
    if os.path.exists(megadetector_temp_base):
         shutil.rmtree(megadetector_temp_base)
    assert not os.path.exists(megadetector_temp_base)

    # Restore original tempdir if it was changed (though not expected for these functions)
    tempfile.tempdir = original_tempdir


def run_all_module_tests():
    """
    Run all tests in the ct_utils module.  This is not invoked by pytest; this is
    just a convenience wrapper for debugging the tests.
    """

    test_write_json()
    test_path_operations()
    test_geometric_operations()
    test_dictionary_operations()
    test_float_rounding_and_truncation()
    test_object_conversion_and_presentation()
    test_list_operations()
    test_datetime_serialization()
    test_bounding_box_operations()
    test_detection_processing()
    test_type_checking_and_validation()
    test_string_parsing()
    test_temp_folder_creation()



"""

visualize_video_output.py

Render a folder of videos with bounding boxes to a new folder, based on a
detector output file.

"""

#%% Imports

import argparse
import os
import random
import cv2

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial
from tqdm import tqdm
from PIL import Image
import numpy as np

from megadetector.data_management.annotations.annotation_constants import detector_bbox_category_id_to_name
from megadetector.detection.video_utils import run_callback_on_frames, default_fourcc, is_video_file
from megadetector.utils.path_utils import path_is_abs
from megadetector.utils.path_utils import clean_filename
from megadetector.utils.path_utils import insert_before_extension
from megadetector.utils.wi_taxonomy_utils import load_md_or_speciesnet_file
from megadetector.visualization.visualization_utils import render_detection_bounding_boxes


#%% Constants

# This will only be used if a category mapping is not available in the results file
DEFAULT_DETECTOR_LABEL_MAP = {
    str(k): v for k, v in detector_bbox_category_id_to_name.items()
}

DEFAULT_CLASSIFICATION_THRESHOLD = 0.4
DEFAULT_DETECTION_THRESHOLD = 0.15

NO_FRAMES_STRING = 'No frames with detections to process'
SKIPPED_STRING = 'Skipped'


#%% Classes

class VideoVisualizationOptions:
    """
    Options controlling the behavior of visualize_video_output()
    """

    def __init__(self):

        #: Confidence threshold for including detections
        self.confidence_threshold = DEFAULT_DETECTION_THRESHOLD

        #: Sample N videos to process (-1 for all videos)
        self.sample = -1

        #: Random seed for sampling
        self.random_seed = None

        #: Confidence threshold for including classifications
        self.classification_confidence_threshold = DEFAULT_CLASSIFICATION_THRESHOLD

        #: Frame rate for output videos. Either a float (fps) or 'auto' to calculate
        #: based on detection frame intervals
        self.rendering_fs = 'auto'

        #: Fourcc codec specification for video encoding
        self.fourcc = default_fourcc

        #: Skip frames before first and after last above-threshold detection
        self.trim_to_detections = False

        #: By default, output videos use the same extension as input videos,
        #: use this to force a particular extension
        self.output_extension = None # 'mp4'

        #: By default, relative paths are preserved in the output folder; this
        #: flattens the output structure.
        self.flatten_output = False

        #: When flatten_output is True, path separators will be replaced with this
        #: string.
        self.path_separator_replacement = '#'

        #: Don't render videos below this length
        self.min_output_length_seconds = None

        #: Enable parallel processing of videos
        self.parallelize_rendering = True

        #: Number of concurrent workers (None = default based on CPU count)
        self.parallelize_rendering_n_cores = 8

        #: Use threads (True) vs processes (False) for parallelization
        self.parallelize_rendering_with_threads = True

        #: Should we include classification category names in the output filenames?
        #: Helps for finding showcase videos.  Can be "start", "end", or None.
        self.include_category_names_in_filenames = None

        #: List of category name strings to skip (e.g. "none", "bear_moose"), or None
        self.exclude_category_name_strings = None

# ...class VideoVisualizationOptions


#%% Support functions

def _get_video_output_framerate(video_entry, original_framerate, rendering_fs='auto'):
    """
    Calculate the appropriate output frame rate for a video based on detection frame numbers.

    Args:
        video_entry (dict): video entry from results file, typically containing the key
            'detections'
        original_framerate (float): original frame rate of the video
        rendering_fs (str or float): 'auto' for automatic calculation, negative float for
            speedup factor, positive float for explicit fps

    Returns:
        float: calculated output frame rate
    """

    if rendering_fs != 'auto':

        if float(rendering_fs) < 0:

            # Negative value means speedup factor
            speedup_factor = abs(float(rendering_fs))
            if ('detections' not in video_entry) or (len(video_entry['detections']) == 0):
                # This is a bit arbitrary, but a reasonable thing to do when we have no basis
                # to determine the output frame rate
                return original_framerate * speedup_factor

            frame_numbers = []
            for detection in video_entry['detections']:
                if 'frame_number' in detection:
                    frame_numbers.append(detection['frame_number'])

            if len(frame_numbers) < 2:
                # This is a bit arbitrary, but a reasonable thing to do when we have no basis
                # to determine the output frame rate
                return original_framerate * speedup_factor

            frame_numbers = sorted(set(frame_numbers))
            first_interval = frame_numbers[1] - frame_numbers[0]

            # Calculate base output frame rate based on first interval, then apply speedup
            base_output_fps = original_framerate / first_interval
            return base_output_fps * speedup_factor

        else:

            # Positive value means explicit fps
            return float(rendering_fs)

        # ...if we're using an explicit/speedup-based frame rate

    # ...if we aren't in "auto" frame rate mode

    # Auto mode
    if 'detections' not in video_entry or len(video_entry['detections']) == 0:
        return original_framerate

    frame_numbers = []
    for detection in video_entry['detections']:
        if 'frame_number' in detection:
            frame_numbers.append(detection['frame_number'])

    frame_numbers = sorted(set(frame_numbers))

    if len(frame_numbers) < 2:
        return original_framerate

    first_interval = frame_numbers[1] - frame_numbers[0]

    # Calculate output frame rate based on first interval
    output_fps = original_framerate / first_interval

    return output_fps


def _get_frames_to_process(video_entry, confidence_threshold, trim_to_detections=False):
    """
    Get list of frame numbers that have detections for this video.

    Args:
        video_entry (dict): video entry from results file
        confidence_threshold (float): minimum confidence for detections to be considered
        trim_to_detections (bool): if True, only include frames between first and last
            above-threshold detections (inclusive)

    Returns:
        list: sorted list of unique frame numbers to process
    """

    if 'detections' not in video_entry:
        return []

    if 'frames_processed' in video_entry:
        frame_numbers = set(video_entry['frames_processed'])
    else:
        frame_numbers = set()

    for detection in video_entry['detections']:

        if 'frame_number' in detection:
            # If this file includes the list of frames processed (required as of format
            # version 1.5), every frame with detections should be included in that list
            if 'frames_processed' in video_entry:
                if detection['frame_number'] not in frame_numbers:
                    print('Warning: frames_processed field present in {}, but frame {} is missing'.\
                          format(video_entry['file'],detection['frame_number']))
            frame_numbers.add(detection['frame_number'])
        else:
            print('Warning: detections in {} lack frame numbers'.format(video_entry['file']))

    # ...for each detection

    frame_numbers = sorted(list(frame_numbers))

    if trim_to_detections and (len(frame_numbers) > 0):

        # Find first and last frames with above-threshold detections

        above_threshold_frames = set()
        for detection in video_entry['detections']:
            if detection['conf'] >= confidence_threshold:
                above_threshold_frames.add(detection['frame_number'])

        if len(above_threshold_frames) > 0:

            above_threshold_frames = sorted(list(above_threshold_frames))
            first_detection_frame = above_threshold_frames[0]
            last_detection_frame = above_threshold_frames[-1]

            # Return all frames between first and last above-threshold detections (inclusive)
            trimmed_frames = []
            for frame_num in frame_numbers:
                if (first_detection_frame <= frame_num) and (frame_num <= last_detection_frame):
                    trimmed_frames.append(frame_num)
            return trimmed_frames

        else:
            # No above-threshold detections, return empty list
            return []

    # ...if we're supposed to be trimming to non-empty frames

    return frame_numbers


def _get_detections_for_frame(video_entry, frame_number, confidence_threshold):
    """
    Get all detections for a specific frame that meet confidence thresholds.

    Args:
        video_entry (dict): video entry from results file
        frame_number (int): frame number to get detections for
        confidence_threshold (float): minimum detection confidence

    Returns:
        list: list of detection dictionaries for this frame
    """

    if 'detections' not in video_entry:
        return []

    frame_detections = []

    for detection in video_entry['detections']:
        if ((detection['frame_number'] == frame_number) and
            (detection['conf'] >= confidence_threshold)):
            frame_detections.append(detection)

    return frame_detections


def _get_classification_names(video_entry,classification_label_map,options):
    """
    Return the set of above-threshold classification category names in [video_entry].

    video_entry (dict): video entry from results file, typically containing keys
            'file', 'detections', and 'frame_rate'
        detector_label_map (dict): mapping of detection category IDs to names
        classification_label_map (dict): mapping of classification category IDs to names
        options (VideoVisualizationOptions): processing options

    Returns:
        set: the above-threshold classification category names for this video,
        an empty set if no above-threshold classifications are present
    """

    classification_names = set()

    if ('detections' not in video_entry) or (video_entry['detections'] is None):
        return classification_names

    for det in video_entry['detections']:

        if det['conf'] < options.confidence_threshold:
            continue

        if 'classifications' not in det:
            continue

        for classification in det['classifications']:

            classification_category_id = classification[0]
            classification_conf = classification[1]
            if classification_conf < options.classification_confidence_threshold:
                continue
            classification_category_name = classification_label_map[classification_category_id]
            classification_category_name = clean_filename(classification_category_name,force_lower=True)
            classification_names.add(classification_category_name)

        # ...for each classification

    # ...for each detection

    return classification_names

# ...def _get_classification_names(...)


def _process_video(video_entry,
                   detector_label_map,
                   classification_label_map,
                   options,
                   video_dir,
                   out_dir):
    """
    Process a single video, rendering detections on frames and creating output video.

    Args:
        video_entry (dict): video entry from results file, typically containing keys
            'file', 'detections', and 'frame_rate'
        detector_label_map (dict): mapping of detection category IDs to names
        classification_label_map (dict): mapping of classification category IDs to names
        options (VideoVisualizationOptions): processing options
        video_dir (str): input video directory
        out_dir (str): output directory

    Returns:
        dict: processing result information, with at least keys 'file, 'error', 'success',
        'frames_processed'.
    """

    result = {
        'file': video_entry['file'],
        'success': False,
        'error': None,
        'frames_processed': 0
    }

    # Handle failure cases
    if ('failure' in video_entry) and (video_entry['failure'] is not None):
        result['error'] = 'Ignoring failed video: {}'.format(video_entry['failure'])
        return result

    # Construct input and output paths
    if video_dir is None:
        input_video_path = video_entry['file']
        assert path_is_abs(input_video_path), \
            'Absolute paths are required when no video base dir is supplied'
    else:
        assert not path_is_abs(video_entry['file']), \
            'Relative paths are required when a video base dir is supplied'
        input_video_path = os.path.join(video_dir, video_entry['file'])

    if not os.path.exists(input_video_path):
        result['error'] = 'Video not found: {}'.format(input_video_path)
        return result

    output_fn_relative = video_entry['file']

    if options.flatten_output:
        output_fn_relative = output_fn_relative.replace('\\','/')
        output_fn_relative = \
            output_fn_relative.replace('/',options.path_separator_replacement)

    if options.output_extension is not None:
        ext = options.output_extension
        if not ext.startswith('.'):
            ext = '.' + ext
        output_fn_relative = os.path.splitext(output_fn_relative)[0] + ext

    category_names = _get_classification_names(video_entry,classification_label_map,options)
    if len(category_names) > 0:
        category_names = '_'.join(sorted(list(category_names)))
    else:
        category_names = 'none'

    if options.exclude_category_name_strings is not None:
        if category_names in options.exclude_category_name_strings:
            print('Ignoring category string {} for {}'.format(
                category_names,output_fn_relative))
            result['error'] = SKIPPED_STRING + ': {}'.format(category_names)
            return result

    if options.include_category_names_in_filenames is not None:
        if options.include_category_names_in_filenames == 'start':
            output_fn_relative = category_names + '_' + output_fn_relative
        else:
            output_fn_relative = insert_before_extension(output_fn_relative,category_names)

    output_fn_abs = os.path.join(out_dir, output_fn_relative)
    parent_dir = os.path.dirname(output_fn_abs)
    if len(parent_dir) > 0:
        os.makedirs(parent_dir, exist_ok=True)

    # Get frames to process
    frames_to_process = _get_frames_to_process(video_entry,
                                               options.confidence_threshold,
                                               options.trim_to_detections)
    if len(frames_to_process) == 0:
        result['error'] = NO_FRAMES_STRING
        return result

    # Determine output frame rate
    original_framerate = video_entry['frame_rate']
    output_framerate = _get_video_output_framerate(video_entry,
                                                   original_framerate,
                                                   options.rendering_fs)

    # Bail early if this video is below the output length limit
    if options.min_output_length_seconds is not None:
        output_length = len(frames_to_process) / output_framerate
        if output_length < options.min_output_length_seconds:
            print('Skipping video {}, {}s is below minimum length ({}s)'.format(
                video_entry['file'],output_length,options.min_output_length_seconds))
            result['error'] = 'Skipped, below minimum length'
            return result

    # Storage for rendered frames
    rendered_frames = []

    def frame_callback(frame_array, frame_id):
        """
        Callback function for processing each frame.

        Args:
            frame_array (np.array): frame image data
            frame_id (str): frame identifier (unused)

        Returns:
            np.array: processed frame
        """

        # Extract frame number from the current processing context
        current_frame_idx = len(rendered_frames)
        if current_frame_idx >= len(frames_to_process):
            print('Warning: received an extra frame (index {} of {}) for video {}'.format(
                current_frame_idx,len(frames_to_process),video_entry['file']
            ))
            return frame_array

        current_frame_number = frames_to_process[current_frame_idx]

        # Convert numpy array to PIL Image
        if frame_array.dtype != np.uint8:
            frame_array = (frame_array * 255).astype(np.uint8)

        # Convert from BGR (OpenCV) to RGB (PIL) if needed
        if len(frame_array.shape) == 3 and frame_array.shape[2] == 3:
            frame_array = cv2.cvtColor(frame_array, cv2.COLOR_BGR2RGB)

        pil_image = Image.fromarray(frame_array)

        # Get detections for this frame
        frame_detections = _get_detections_for_frame(
            video_entry,
            current_frame_number,
            options.confidence_threshold
        )

        # Render detections on the frame
        if frame_detections:
            render_detection_bounding_boxes(
                frame_detections,
                pil_image,
                detector_label_map,
                classification_label_map,
                classification_confidence_threshold=options.classification_confidence_threshold
            )

        # Convert back to numpy array for video writing
        frame_array = np.array(pil_image)
        if (len(frame_array.shape) == 3) and (frame_array.shape[2] == 3):
            frame_array = cv2.cvtColor(frame_array, cv2.COLOR_RGB2BGR)

        rendered_frames.append(frame_array)
        return frame_array

    # ...def frame_callback(...)

    # Process video frames
    try:
        run_callback_on_frames(
            input_video_path,
            frame_callback,
            frames_to_process=frames_to_process,
            verbose=False
        )
    except Exception as e:
        import traceback
        trace = traceback.format_exc()
        result['error'] = 'Error processing video frames: {} ({})'.format(str(e),trace)
        return result

    # Write output video
    if len(rendered_frames) > 0:

        video_writer = None

        try:

            # Get frame dimensions
            height, width = rendered_frames[0].shape[:2]

            # Create VideoWriter
            fourcc = cv2.VideoWriter_fourcc(*options.fourcc)
            video_writer = cv2.VideoWriter(output_fn_abs, fourcc, output_framerate, (width, height))

            if not video_writer.isOpened():
                result['error'] = 'Failed to open video writer for {}'.format(output_fn_abs)
                return result

            # Write frames
            for frame in rendered_frames:
                video_writer.write(frame)

            result['success'] = True
            result['frames_processed'] = len(rendered_frames)

        except Exception as e:

            result['error'] = 'Error writing output video: {}'.format(str(e))
            return result

        finally:

            if video_writer is not None:
                try:
                    video_writer.release()
                except Exception as e:
                    print('Warning: failed to release video writer for file {}: {}'.format(
                        video_entry['file'],str(e)))

        # ...try/except

    else:

        result['error'] = 'No frames were processed for video {}'.format(video_entry['file'])

    return result

# ...def _process_video(...)


#%% Main function

def visualize_video_output(detector_output_path,
                           out_dir,
                           video_dir,
                           options=None):
    """
    Renders videos with bounding boxes based on detector output.

    Args:
        detector_output_path (str): path to .json file containing detection results
        out_dir (str): output directory for rendered videos
        video_dir (str): input video directory
        options (VideoVisualizationOptions, optional): processing options

    Returns:
        list: list of processing results for each video
    """

    if options is None:
        options = VideoVisualizationOptions()

    # Validate that input and output directories are different
    if (video_dir is not None) and (os.path.abspath(out_dir) == os.path.abspath(video_dir)):
        raise ValueError('Output directory cannot be the same as video directory')

    # Load results file
    print('Loading results from {}'.format(detector_output_path))
    results_data = load_md_or_speciesnet_file(detector_output_path)

    # Get label mappings
    detector_label_map = results_data.get('detection_categories', DEFAULT_DETECTOR_LABEL_MAP)
    classification_label_map = results_data.get('classification_categories', {})

    # Filter to video entries only
    video_entries = []
    for entry in results_data['images']:
        if is_video_file(entry['file']):
            video_entries.append(entry)

    print('Found {} videos in results file'.format(len(video_entries)))

    # Apply sampling if requested
    if (options.sample > 0) and (len(video_entries) > options.sample):
        if options.random_seed is not None:
            random.seed(options.random_seed)
        n_videos_available = len(video_entries)
        video_entries = random.sample(video_entries, options.sample)
        print('Sampled {} of {} videos for processing'.format(
            len(video_entries),n_videos_available))

    # Create output directory
    os.makedirs(out_dir, exist_ok=True)

    # Process each video
    results = []

    if options.parallelize_rendering:

        if options.parallelize_rendering_with_threads:
            worker_string = 'threads'
        else:
            worker_string = 'processes'

        pool = None

        try:

            if options.parallelize_rendering_n_cores is None:
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool()
                else:
                    pool = Pool()
            else:
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool(options.parallelize_rendering_n_cores)
                else:
                    pool = Pool(options.parallelize_rendering_n_cores)
                print('Processing videos with {} {}'.format(options.parallelize_rendering_n_cores,
                                                            worker_string))
            results = list(tqdm(pool.imap(
                                 partial(_process_video,
                                         detector_label_map=detector_label_map,
                                         classification_label_map=classification_label_map,
                                         options=options,
                                         video_dir=video_dir,
                                         out_dir=out_dir),
                                 video_entries), total=len(video_entries), desc='Processing videos'))
        finally:

            if pool is not None:
                pool.close()
                pool.join()
                print('Pool closed and joined for video output visualization')

    else:

        for video_entry in tqdm(video_entries, desc='Processing videos'):

            result = _process_video(
                video_entry,
                detector_label_map,
                classification_label_map,
                options,
                video_dir,
                out_dir
            )
            results.append(result)

    # ...for each video

    # Print summary
    n_empty = 0
    n_failed = 0
    n_successful = 0
    n_skipped = 0

    for r in results:
        if r['success']:
            assert r['error'] is None
            n_successful += 1
        else:
            assert r['error'] is not None
            if NO_FRAMES_STRING in r['error']:
                n_empty += 1
            elif SKIPPED_STRING in r['error']:
                n_skipped += 1
            else:
                n_failed += 1

    total_frames = sum(r['frames_processed'] for r in results if r['success'])

    print('\nProcessing complete:')
    print(f'  Successfully processed: {n_successful} videos')
    print(f'  No above-threshold detections: {n_empty} videos')
    print(f'  Failed: {n_failed} videos')
    print(f'  Skipped: {n_skipped} videos')
    print(f'  Total frames rendered: {total_frames}')

    return results

# ...def visualize_video_output(...)


#%% Command-line driver

def main():
    """
    Command-line driver for visualize_video_output
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Render videos with bounding boxes predicted by a detector above '
                    'a confidence threshold, and save the rendered videos.')

    parser.add_argument(
        'detector_output_path',
        type=str,
        help='Path to json output file of the detector')

    parser.add_argument(
        'out_dir',
        type=str,
        help='Path to directory where the rendered videos will be saved. '
             'The directory will be created if it does not exist.')

    parser.add_argument(
        'video_dir',
        type=str,
        help='Path to directory containing the input videos')

    parser.add_argument(
        '--confidence_threshold',
        type=float,
        default=DEFAULT_DETECTION_THRESHOLD,
        help='Confidence threshold above which detections will be rendered')

    parser.add_argument(
        '--sample',
        type=int,
        default=-1,
        help='Number of videos to randomly sample for processing. '
             'Set to -1 to process all videos')

    parser.add_argument(
        '--random_seed',
        type=int,
        default=None,
        help='Random seed for reproducible sampling')

    parser.add_argument(
        '--classification_confidence_threshold',
        type=float,
        default=DEFAULT_CLASSIFICATION_THRESHOLD,
        help='Value between 0 and 1, indicating the confidence threshold '
             'above which classifications will be rendered')

    parser.add_argument(
        '--rendering_fs',
        default='auto',
        help='Frame rate for output videos. Use "auto" to calculate based on '
             'detection frame intervals, positive float for explicit fps, '
             'or negative float for speedup factor (e.g. -2.0 = 2x faster)')

    parser.add_argument(
        '--fourcc',
        type=str,
        default=default_fourcc,
        help='Fourcc codec specification for video encoding')

    parser.add_argument(
        '--trim_to_detections',
        action='store_true',
        help='Skip frames before first and after last above-threshold detection')

    args = parser.parse_args()

    # Create options object
    options = VideoVisualizationOptions()
    options.confidence_threshold = args.confidence_threshold
    options.sample = args.sample
    options.random_seed = args.random_seed
    options.classification_confidence_threshold = args.classification_confidence_threshold
    options.rendering_fs = args.rendering_fs
    options.fourcc = args.fourcc
    options.trim_to_detections = args.trim_to_detections

    # Run visualization
    visualize_video_output(
        args.detector_output_path,
        args.out_dir,
        args.video_dir,
        options
    )


if __name__ == '__main__':
    main()
